"""SQLAlchemy goodies that provides a nice interface to using sqlalchemy with
django."""
import functools
import inspect

import sqlalchemy as sa
import sqlalchemy.orm  # noqa
from sqlalchemy.ext.declarative import declarative_base

from ..utils import make_args
from . import fields, signals
from .composites import BaseComposite, CompositeField
from .models import Base, BaseMeta
from .query import Query, QueryProperty
from .relations import RelationsMixin
from .session import SignallingSession
from .transaction import TransactionContext


def instrument(name):
    def do(self, *args, **kwargs):
        return getattr(self.registry(), name)(*args, **kwargs)

    return do


def makeprop(name):
    def get(self):
        return getattr(self.registry(), name)

    return property(get)


class _sqla_meta(type):
    def __new__(mcs, name, bases, attrs):
        typ = super().__new__(mcs, name, bases, attrs)

        # figure out all props to be proxied
        dummy = sa.orm.Session()
        props = {i for i in dir(dummy) if not i.startswith("__")}
        props.update(
            [
                "__contains__",
                "__iter__",
                "add",
                "add_all",
                "autocommit",
                "autoflush",
                "begin",
                "begin_nested",
                "bind",
                "bulk_insert_mappings",
                "bulk_save_objects",
                "bulk_update_mappings",
                "close",
                "commit",
                "connection",
                "delete",
                "deleted",
                "dirty",
                "execute",
                "expire",
                "expire_all",
                "expunge",
                "expunge_all",
                "flush",
                "get_bind",
                "identity_map",
                "info",
                "is_active",
                "is_modified",
                "merge",
                "new",
                "no_autoflush",
                "query",
                "refresh",
                "rollback",
                "scalar",
            ]
        )

        for i in props:
            if not hasattr(typ, name):
                if hasattr(dummy, i) and inspect.isroutine(getattr(dummy, i)):
                    setattr(typ, i, instrument(i))
                else:
                    setattr(typ, i, makeprop(i))

        return typ


class SQLAlchemy(RelationsMixin, metaclass=_sqla_meta):
    """This class itself is a scoped session and provides very thin and useful
    abstractions and conventions for using sqlalchemy with django."""

    session_class = SignallingSession
    query_class = Query
    registry_class = sa.util.ThreadLocalRegistry
    metadata_class = sa.MetaData
    model_class = Base

    BaseComposite = BaseComposite
    CompositeField = CompositeField

    def __init__(self, url, **kwargs):
        self.url, self.kwargs = url, kwargs
        self.alias = kwargs.get("alias")
        self.session_class = self.kwargs.get("session_class", None) or self.session_class
        self.query_class = self.kwargs.get("query_class", None) or self.query_class
        self.registry_class = self.kwargs.get("registry_class", None) or self.registry_class
        self.metadata_class = self.kwargs.get("metadata_class", None) or self.metadata_class
        self.model_class = self.kwargs.get("model_class", None) or self.model_class
        self.engine_options = self.kwargs.get("engine_options", {})
        self.session_options = self.kwargs.get("session_options", {})

        self.session_options.setdefault("query_cls", self.query_class)
        self.session_options.setdefault("class_", self.session_class)

        self.middleware = self.make_middleware()
        self.models_registry = []
        self.metadata = self.metadata_class(**self.kwargs.get("metadata_kwargs", {}))
        self.Model = self._make_declarative(self.model_class)

        for module, is_partial in [(sa, False), (sa.sql, False), (sa.orm, False), (fields, True)]:
            for key in module.__all__:
                if not hasattr(self, key) and not hasattr(type(self), key):
                    value = getattr(module, key)
                    setattr(
                        self, key, functools.wraps(value)(functools.partial(value, db=self)) if is_partial else value
                    )

        self.collections = sa.orm.collections
        self.event = sa.event
        self.relationship = self._wrap(self.relationship)
        self.relation = self._wrap(self.relation)
        self.dynamic_loader = self._wrap(self.dynamic_loader)
        self._registry = None

    def __call__(self, **kwargs):
        return self.session(**kwargs)

    @property
    def registry(self):
        """Returns scoped registry instance."""
        if not self._registry:
            engine = self._create_engine(self.url, **self.engine_options)
            self._registry = self.registry_class(sa.orm.sessionmaker(bind=engine, **self.session_options))

        return self._registry

    @property
    def inspector(self):
        """Returns engine inspector.

        Useful for querying for db schema info.
        """
        return sa.inspect(self.engine)

    def session(self, **kwargs):
        """Return the current session, creating it if necessary using
        session_factory for the current scope Any kwargs provided will be
        passed on to the session_factory.

        If there's already a session in current scope, will raise
        InvalidRequestError
        """
        if not kwargs:
            return self.registry()

        if self.registry.has():
            raise sa.exc.InvalidRequestError("Scoped session is already present; " "no new arguments may be specified.")

        session = self.session_factory(**kwargs)
        self.registry.set(session)
        return session

    def Table(self, name, *args, **kwargs):
        """Returns a sqlalchemy table that is automatically added to
        metadata."""
        assert name is not None, "Table requires `name` argument"
        assert args, "Table at least takes one column argument"
        for arg in args:
            assert not isinstance(arg, sa.MetaData), "Passing a `metadata` is not allowed"

        return sa.Table(name, self.metadata, *args, **kwargs)

    def _set_query_class(self, cls, kwargs):
        query_class = self.query_class
        try:
            query_class = cls.query_class
        except AttributeError:
            pass

        kwargs["query_class"] = query_class

    def _wrap(self, fn):
        """Wrapper that sets the 'query_class' argument from the model's
        query_class attribute."""

        @functools.wraps(fn)
        def func(cls, *args, **kwargs):
            """Wraps a function setting default `query_class` argument."""
            self._set_query_class(cls, kwargs)

            if "backref" in kwargs:
                backref = kwargs["backref"]
                backref_kwargs = {}
                if isinstance(backref, tuple):
                    backref, backref_kwargs = backref
                    self._set_query_class(backref, backref_kwargs)

                kwargs["backref"] = (backref, backref_kwargs)

            return fn(cls, *args, **kwargs)

        return func

    def _create_engine(self, url, **kwargs):
        engine = sa.create_engine(url, **kwargs)
        signals.engine_created.send(engine)
        return engine

    def _make_declarative(self, model):
        """
        Creates the base class that the models should inherit from
        ----------------------------------------------------------
        model: class
            The base class for the declarative_base to be inherited from
        """
        base = declarative_base(
            cls=model, metadata=self.metadata, metaclass=type("BaseMeta", (BaseMeta,), {"db": self})
        )

        # allow to customize things in custom base model
        if not hasattr(base, "query_class"):
            base.query_class = self.query_class
        if not hasattr(base, "query"):
            base.query = self.queryproperty()
        if not hasattr(base, "objects"):
            base.objects = self.queryproperty()

        return base

    @property
    def engine(self):
        """Current engine."""
        return self.bind

    @property
    def session_factory(self):
        """Current session factory to create sessions."""
        return self.registry.createfunc

    def __repr__(self):
        return "<{} engine={!r}>".format(self.__class__.__name__, self.url)

    def queryproperty(self, *args, **kwargs):
        """Generate a query property for a model."""
        model = None
        if args and isinstance(args[0], type) and issubclass(args[0], self.Model):
            model = args[0]
            args = args[1:]

        return QueryProperty(self, model, *args, **kwargs)

    def atomic(self, savepoint=True):
        """Create a savepoint or transaction scope."""
        return TransactionContext(self, savepoint=savepoint)

    def make_middleware(self):
        """Creates a middleware to be used in a django application."""
        from .middleware import SQLAlchemyDBMiddleware

        return type("SQLAlchemyMiddleware", (SQLAlchemyDBMiddleware,), {"db": self})

    def args(self, *args, **kwargs):
        """Useful for setting table args and mapper args on models."""
        return make_args(*args, **kwargs)

    def remove(self):
        """Remove the current scoped session."""
        if self.registry.has():
            self.registry().close()
        self.registry.clear()
        for signal in signals.all_signals.scoped_signals:
            signal.cleanup()

    def create_all(self):
        """Create the schema in db."""
        result = signals.before_create_all.send(db=self, bind=self.engine)
        if all(i[1] in [True, None] for i in result):
            self.metadata.create_all(bind=self.engine)
            signals.after_create_all.send(self, db=self, bind=self.engine)

    def drop_all(self):
        """Drop the schema in db."""
        result = signals.before_drop_all.send(db=self, bind=self.engine)
        if all(i[1] in [True, None] for i in result):
            self.metadata.drop_all(bind=self.engine)
            signals.after_drop_all.send(self, db=self, bind=self.engine)



"""sqlalchemy model related things."""
from itertools import chain

import sqlalchemy as sa
import sqlalchemy.ext.declarative  # noqa
import sqlalchemy.orm  # noqa
from django.core.exceptions import ValidationError
from django.utils.text import camel_case_to_spaces
from sqlalchemy.orm.base import MANYTOONE, NO_VALUE

from . import meta, signals
from .mixins import CleanMixin


def simple_repr(instance, fields=None):
    """
    Provides basic repr for models
    ------------------------------
    instance: Model
        a model instance
    fields: list
        list of fields to display on repr
    """
    info = meta.model_info(instance)
    state = info.sa_state(instance)
    pks = info.primary_keys
    fields = fields or info.properties

    pk_reprs = []
    for key in pks:
        value = state.attrs[key].loaded_value
        if value == NO_VALUE:
            value = None
        elif isinstance(value, str):
            # remove pesky "u" prefix in repr for strings
            value = str(value)
        pk_reprs.append("=".join([key, repr(value)]))

    field_reprs = []
    for key in fields:
        value = state.attrs[key].loaded_value
        if isinstance(value, str):
            # remove pesky "u" prefix in repr for strings
            value = str(value)
        if value != NO_VALUE:
            field_reprs.append("=".join([key, repr(value)]))

    return "".join([instance.__class__.__name__, "(", ", ".join(chain(pk_reprs, sorted(field_reprs))), ")"])


def serialize(instance, *rels):
    """
    Return a dict of column attributes
    ------------------------------
    instance:
        a model instance
    rels: list of relations
        relationships to be serialized
    """
    if instance is None:
        return None

    if isinstance(instance, (list, set)):
        return [serialize(i, *rels) for i in instance]

    info = meta.model_info(instance)
    rels = set(rels)

    data = {name: getattr(instance, name) for name, _ in info.column_properties}

    for name, composite in info.composites.items():
        comp = getattr(instance, name)
        data[name] = vars(comp) if comp else None
        # since we're copying, remove props from the composite
        for _, prop in composite.properties.items():
            data.pop(prop.property.key, None)

    for name in info.relationships:
        attr = getattr(info.model_class, name)
        if attr in rels:
            sub_instance = getattr(instance, name, None)
            sub_rels = [r for r in rels if r is not attr]
            data[name] = serialize(sub_instance, *sub_rels)

    return data


def deserialize(model, data):
    """
    Return a model instance from data
    ------------------------------
    model:
        a model class
    data: dict
        values
    """
    identity_map = {}
    instance = _deserialize(model, data, identity_map)

    for val in identity_map.values():
        info = meta.model_info(val.__class__)
        for prop, rel in info.relationships.items():
            if rel.direction == MANYTOONE or not rel.uselist:
                deserialized_instance = getattr(val, prop)
                if deserialized_instance is not None:
                    continue

                fks = {}
                for local, remote in rel.local_remote_pairs_for_identity_key:
                    local_attr = rel.parent_mapper.get_property_by_column(local)
                    remote_attr = rel.related_mapper.get_property_by_column(remote)
                    fks[remote_attr.key] = getattr(val, local_attr.key)

                related_info = meta.model_info(rel.related_model)
                ident_key = related_info.identity_key_from_dict(fks)
                if ident_key is not None and ident_key in identity_map:
                    setattr(val, prop, identity_map[ident_key])

    return instance


def _deserialize(model, data, identity_map):
    if data is None:
        return None

    if isinstance(data, (list, tuple, set)):
        return [_deserialize(model, i, identity_map) for i in data]

    info = meta.model_info(model)

    kwargs = {prop: data.get(prop) for prop in info.primary_keys if prop in data}
    pk = info.identity_key_from_dict(kwargs)
    if pk is not None and pk in identity_map:
        return identity_map[pk]

    for prop in info.properties:
        if prop in data:
            kwargs[prop] = data.get(prop)

    for prop, composite in info.composites.items():
        if prop in data:
            composite_data = data.get(prop)
            composite_class = composite.model_class
            composite_args = [composite_data.get(i) for i in composite.properties]
            kwargs[prop] = composite_class(*composite_args)

    for prop, rel in info.relationships.items():
        if prop in data:
            kwargs[prop] = _deserialize(rel.related_model, data.get(prop), identity_map)

    instance = model(**kwargs)

    if pk is not None:
        identity_map[pk] = instance

    return instance


def clone(instance, *rels, **kwargs):
    """
    Clone a model instance with or without any relationships
    --------------------------------------------------------
    instance: Model
        a model instance
    relations: list or relations or a tuple of relation and kwargs for that relation
        relationships to be cloned with relationship and optionally kwargs
    kwargs: dict string of any
        attribute values to be overridden
    """
    if instance is None:
        return None

    if isinstance(instance, (list, set)):
        return [clone(i, *rels, **kwargs) for i in instance]

    relations = {}
    for rel in rels:
        r = rel
        rkwargs = {}
        if isinstance(rel, tuple):
            r, rkwargs = rel

        relations[r] = rkwargs

    mapper = sa.inspect(instance).mapper
    kwargs = kwargs or {}

    fks = set(
        chain(
            *[
                fk.columns
                for fk in chain(*[table.constraints for table in mapper.tables])
                if isinstance(fk, sa.ForeignKeyConstraint)
            ]
        )
    )

    for column in mapper.columns:
        prop = mapper.get_property_by_column(column)

        if prop.key in kwargs:
            continue

        if column.primary_key or column in fks:
            continue

        kwargs[prop.key] = getattr(instance, prop.key)

    for composite in mapper.composites:
        value = getattr(instance, composite.key, None)
        if value:
            kwargs[composite.key] = composite.composite_class(*value.__composite_values__())
            # since we're copying, remove props from the composite
            for prop in composite.props:
                kwargs.pop(prop.key, None)

    for relation in mapper.relationships:
        if relation.key in kwargs:
            continue

        attr = getattr(mapper.class_, relation.key)
        if attr in relations:
            sub_instance = getattr(instance, relation.key, None)
            sub_rels = [(r, kw) for r, kw in relations.items() if r is not attr]
            kwargs[relation.key] = clone(sub_instance, *sub_rels, **relations[attr])

    return mapper.class_(**kwargs)


class BaseMeta(sqlalchemy.ext.declarative.DeclarativeMeta):
    """Base metaclass for models which registers models to DB model registry
    when models are created."""

    def __new__(mcs, name, bases, attrs):
        klass = super().__new__(mcs, name, bases, attrs)
        mcs.db.models_registry.append(klass)
        return klass


def _table_name(cls):
    opts = getattr(cls, "Meta", None)
    return getattr(opts, "db_table", None) or "_".join(camel_case_to_spaces(cls.__name__).split())


def _table_args(cls):
    opts = getattr(cls, "Meta", None)
    return getattr(opts, "table_args", ())


def _mapper_args(cls):
    opts = getattr(cls, "Meta", None)
    return getattr(opts, "mapper_args", ())


class Base(CleanMixin):
    """Base model class for SQLAlchemy.

    Can be overwritten by subclasses:

        query_class = None

    Automatically added by declarative base for easier querying:

        query = None
        objects = None
    """

    __tablename__ = sa.ext.declarative.declared_attr(_table_name)
    __table_args__ = sa.ext.declarative.declared_attr(_table_args)
    __mapper_args__ = sa.ext.declarative.declared_attr(_mapper_args)

    def as_dict(self):
        """Return a dict of column attributes."""
        return serialize(self)

    def __repr__(self):
        return simple_repr(self)

    @classmethod
    def __declare_first__(cls):
        """Declarative hook called within `before_configure` mapper event, can
        be called multiple times."""
        signals.declare_first.send(cls)

    @classmethod
    def __declare_last__(cls):
        """Declarative hook called within `after_configure` mapper event, can
        be called multiple times."""
        signals.declare_last.send(cls)

    def _get_properties_for_validation(self):
        """Return all model columns which can be validated."""
        return meta.model_info(self.__class__).properties

    def _get_nested_objects_for_validation(self):
        """Return all composites to be validated."""
        return meta.model_info(self.__class__).composites

    def _get_relation_objects_for_validation(self):
        """Return all relations to be validated."""
        return meta.model_info(self.__class__).relationships


_instant_defaults = set()


def instant_defaults(cls):
    """
    This function automatically registers attribute events that sets the column defaults to a
    model instance at model instance initialization provided that default values are simple
    types::

        @instant_defaults
        class MyModel(db.Model):
            attr = db.Column(..., default=1)

        assert MyModel().default == 1

    """
    mapper = sa.inspect(cls, raiseerr=False)
    if not mapper:
        return

    _instant_defaults.add(cls)
    return cls


signals.declare_last.connect(instant_defaults)


def _set_instant_defaults(target, args, kwargs):
    info = meta.model_info(target.__class__)
    for prop in info.properties.values():
        if not prop.column.default or not hasattr(prop.column.default, "arg") or callable(prop.column.default.arg):
            continue  # pragma: nocover

        if getattr(target, prop.name, None) is None:
            setattr(target, prop.name, prop.column.default.arg)


@sa.event.listens_for(sa.orm.mapper, "after_configured")
def _configure_instant_defaults():
    for cls in _instant_defaults:
        if not sa.event.contains(cls, "init", _set_instant_defaults):
            sa.event.listen(cls, "init", _set_instant_defaults)

    _instant_defaults.clear()


@signals.before_flush.connect
def full_clean_flush_handler(session, **kwargs):
    """Signal handler for executing ``full_clean`` on all dirty and new objects
    in session."""
    for i in session.dirty | session.new:
        if isinstance(i, Base):
            i.full_clean()


_autocoerce_attrs = set()


def autocoerce_properties(*attrs):
    """
    This function automatically registers attribute events that coerces types for given attributes
    using django's form fields.

    ::
        class MyModel(db.Model):
            field1 = Column(...)
            field2 = Column(...)
            field3 = Column(...)
            ...

        autocoerce_properties(MyModel.field1, MyModel.field2)  # Will only force autocoersion on field1 and field2
    """
    _autocoerce_attrs.update(attrs)


def autocoerce(cls):
    """
    This function automatically registers attribute events that coerces types for the attribute
    using django's form fields for a given model classs. If no class is provided, it will wire up
    coersion for all mappers so it can be used as a class decorator or globally.

    ::

        @autocoerce_properties
        class MyModel(db.Model):
            ...

    or::

        class MyModel(db.Model):
            ...

        autocoerce_properties()

    Since django form fields are used for coersion, localization settings such as `USE_THOUSAND_SEPARATOR`,
    `DATE_INPUT_FORMATS` and `DATETIME_INPUT_FORMATS` control type conversions.
    """
    m = sa.inspect(cls)
    autocoerce_properties(*[getattr(cls, attr.key) for attr in m.column_attrs])
    return cls


def _coerce(target, value, oldvalue, initiator):
    minfo = meta.model_info(target)
    cinfo = minfo.properties.get(initiator.key) or minfo.primary_keys.get(initiator.key)
    try:
        return cinfo.to_python(value) if cinfo else value
    except ValidationError as ex:
        raise ValidationError({initiator.key: ex})


@sa.event.listens_for(sa.orm.mapper, "after_configured")
def _configure_coercers():
    for target in _autocoerce_attrs:
        if not sa.event.contains(target, "set", _coerce):
            sa.event.listen(target, "set", _coerce, retval=True)

    _autocoerce_attrs.clear()



"""Django list view things for sqlalchemy."""

from django.core.exceptions import ImproperlyConfigured
from django.views.generic.base import TemplateResponseMixin, View

from .base import BaseMultipleObjectMixin


class MultipleObjectMixin(BaseMultipleObjectMixin):
    """A mixin for views manipulating multiple objects."""

    context_object_name = "object_list"

    def get_context_object_name(self, object_list):
        """Get the name to use for the object."""
        if self.context_object_name:
            return self.context_object_name

        model = self.get_model()
        return "%s_list" % model.__name__.lower()

    def get_context_data(self, object_list=None, **kwargs):
        """Get the context for this view."""
        queryset = object_list if object_list is not None else self.object_list
        page_size = self.get_paginate_by(queryset)
        context_object_name = self.get_context_object_name(queryset)
        if page_size:
            paginator, page, queryset, is_paginated = self.paginate_queryset(queryset, page_size)
            context = {"paginator": paginator, "page_obj": page, "is_paginated": is_paginated, "object_list": queryset}
        else:
            context = {"paginator": None, "page_obj": None, "is_paginated": False, "object_list": queryset}
        if context_object_name is not None:
            context[context_object_name] = queryset
        context.update(kwargs)
        return super().get_context_data(**context)


class BaseListView(MultipleObjectMixin, View):
    """A base view for displaying a list of objects."""

    def get(self, request, *args, **kwargs):
        """Handle GET request on list view."""
        self.object_list = self.get_queryset()
        context = self.get_context_data()
        return self.render_to_response(context)


class MultipleObjectTemplateResponseMixin(TemplateResponseMixin):
    """Mixin for responding with a template and list of objects."""

    template_name_suffix = "_list"

    def get_template_names(self):
        """Return a list of template names to be used for the request.

        Must return a list. May not be called if render_to_response is
        overridden.
        """
        try:
            names = super().get_template_names()
        except ImproperlyConfigured:
            # If template_name isn't specified, it's not a problem --
            # we just start with an empty list.
            names = []

            # If the list is a queryset, we'll invent a template name based on the
            # app and model name. This name gets put at the end of the template
            # name list so that user-supplied names override the automatically-
            # generated ones.

            if hasattr(self, "get_model_template_name"):
                names.append(self.get_model_template_name())
            elif not names:
                raise ImproperlyConfigured(
                    "%(cls)s requires either a 'template_name' attribute or a get_queryset() method that returns a "
                    "QuerySet." % {"cls": self.__class__.__name__}
                )

        return names


class ListView(MultipleObjectTemplateResponseMixin, BaseListView):
    """Render some list of objects, set by `self.model` or `self.queryset`.

    `self.queryset` can actually be any iterable of items, not just a
    queryset.
    """



import pytest
from django_sorcery.pytest_plugin import sqlalchemy_profiler, transact  # noqa
from django_sorcery.testing import CommitException

from .testapp.models import Business, Owner, db


def test_profiler(sqlalchemy_profiler):  # noqa
    sqlalchemy_profiler.exclude = ["owner"]

    with sqlalchemy_profiler:
        db.add(Owner(first_name="foo", last_name="bar"))
        db.flush()

        assert sqlalchemy_profiler.stats["select"] == 0
        assert sqlalchemy_profiler.stats["insert"] == 0
        assert sqlalchemy_profiler.stats["duration"] >= 0

    with sqlalchemy_profiler:
        b = Business(name="Google")
        db.add(b)
        db.flush()

        assert sqlalchemy_profiler.stats["select"] == 0
        assert sqlalchemy_profiler.stats["insert"] == 1
        assert sqlalchemy_profiler.stats["duration"] > 0

    with sqlalchemy_profiler:
        Business.query.filter_by(id=b.id).one()

        assert sqlalchemy_profiler.stats["select"] == 1
        assert sqlalchemy_profiler.stats["insert"] == 0
        assert sqlalchemy_profiler.stats["duration"] > 0

    db.rollback()
    db.remove()


def test_transact(transact):  # noqa
    db.add(Owner(first_name="foo", last_name="bar"))
    db.flush()

    assert Owner.objects.count() == 1

    with pytest.raises(CommitException):
        db.commit()

    transact.stop()

    assert Owner.objects.count() == 0



"""Django edit view things for sqlalchemy."""

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.views.generic.base import TemplateResponseMixin, View
from django.views.generic.edit import FormMixin

from .. import forms
from .detail import (
    BaseDetailView,
    SingleObjectMixin,
    SingleObjectTemplateResponseMixin,
)


class ModelFormMixin(FormMixin, SingleObjectMixin):
    """Provide a way to show and handle a ModelForm in a request."""

    fields = None

    def get_form_class(self):
        if self.fields is not None and self.form_class:
            raise ImproperlyConfigured("Specifying both 'fields' and 'form_class' is not permitted.")

        if self.form_class:
            return self.form_class

        model = self.get_model()
        return forms.modelform_factory(model, fields=self.fields, session=self.session)

    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        kwargs = super().get_form_kwargs()
        if hasattr(self, "object"):
            kwargs.update({"instance": self.object})

        kwargs["session"] = self.session
        return kwargs

    def get_success_url(self):
        """Return the URL to redirect to after processing a valid form."""
        if self.success_url:
            return self.success_url.format(**vars(self.object))

        raise ImproperlyConfigured(
            "No URL to redirect to. Either provide a url or override this function to return a url"
        )

    def form_valid(self, form):
        self.object = form.save()
        return super().form_valid(form)


class ProcessFormView(View):
    """Render a form on GET and processes it on POST."""

    def get(self, request, *args, **kwargs):
        """Handle GET requests: instantiate a blank version of the form."""
        return self.render_to_response(self.get_context_data())

    def post(self, request, *args, **kwargs):
        """Handle POST requests: instantiate a form instance with the passed
        POST variables and then check if it's valid."""
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)

        return self.form_invalid(form)

    put = post


class BaseFormView(FormMixin, ProcessFormView):
    """A base view for displaying a form."""


class FormView(TemplateResponseMixin, BaseFormView):
    """A view for displaying a form and rendering a template response."""


class BaseCreateView(ModelFormMixin, ProcessFormView):
    """Base view for creating a new object instance.

    Using this base class requires subclassing to provide a response
    mixin.
    """

    def get(self, request, *args, **kwargs):
        self.object = None
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = None
        return super().post(request, *args, **kwargs)


class CreateView(SingleObjectTemplateResponseMixin, BaseCreateView):
    """View for creating a new object, with a response rendered by a
    template."""

    template_name_suffix = "_form"


class BaseUpdateView(ModelFormMixin, ProcessFormView):
    """Base view for updating an existing object.

    Using this base class requires subclassing to provide a response
    mixin.
    """

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)


class UpdateView(SingleObjectTemplateResponseMixin, BaseUpdateView):
    """View for updating an object, with a response rendered by a template."""

    template_name_suffix = "_form"


class DeletionMixin:
    """Provide the ability to delete objects."""

    success_url = None

    def delete(self, request, *args, **kwargs):
        """Call the delete() method on the fetched object and then redirect to
        the success URL."""
        self.object = self.get_object()
        success_url = self.get_success_url()
        self.session.delete(self.object)
        self.session.flush()
        return HttpResponseRedirect(success_url)

    # Add support for browsers which only accept GET and POST for now.

    def post(self, request, *args, **kwargs):
        """Handle POST request on deleve view."""
        return self.delete(request, *args, **kwargs)

    def get_success_url(self):
        """Returns the URL to redirect to after processing deletion."""
        if self.success_url:
            return self.success_url.format(**self.object.__dict__)

        else:
            raise ImproperlyConfigured("No URL to redirect to. Provide a success_url.")


class BaseDeleteView(DeletionMixin, BaseDetailView):
    """Base view for deleting an object.

    Using this base class requires subclassing to provide a response
    mixin.
    """


class DeleteView(SingleObjectTemplateResponseMixin, BaseDeleteView):
    """View for deleting an object retrieved with self.get_object(), with a
    response rendered by a template."""

    template_name_suffix = "_confirm_delete"



import alembic.config
from django.test import TestCase
from django_sorcery.db.alembic.base import setup_config


class TestSignals(TestCase):
    def test_alembic_config(self):
        config = alembic.config.Config()
        setup_config(config)

        self.assertEqual(config.get_section_option("loggers", "keys"), "root,sqlalchemy,alembic")
        self.assertEqual(config.get_section_option("handlers", "keys"), "console")
        self.assertEqual(config.get_section_option("formatters", "keys"), "generic")
        self.assertEqual(config.get_section_option("logger_root", "level"), "WARN")
        self.assertEqual(config.get_section_option("logger_root", "handlers"), "console")
        self.assertEqual(config.get_section_option("logger_sqlalchemy", "level"), "WARN")
        self.assertEqual(config.get_section_option("logger_sqlalchemy", "qualname"), "sqlalchemy.engine")
        self.assertEqual(config.get_section_option("logger_alembic", "level"), "INFO")
        self.assertEqual(config.get_section_option("logger_alembic", "qualname"), "alembic")
        self.assertEqual(config.get_section_option("handler_console", "class"), "StreamHandler")
        self.assertEqual(config.get_section_option("handler_console", "formatter"), "generic")



"""Some common utilities."""
import inspect
import unicodedata

from django.conf import settings


def sanitize_separators(value):
    """Sanitize a value according to the current decimal and thousand separator
    setting.

    Used with form field input.
    """
    if isinstance(value, str):
        parts = []
        decimal_separator = settings.DECIMAL_SEPARATOR
        if decimal_separator in value:
            value, decimals = value.split(decimal_separator, 1)
            parts.append(decimals)
        thousand_sep = settings.THOUSAND_SEPARATOR
        for replacement in {thousand_sep, unicodedata.normalize("NFKD", thousand_sep)}:
            value = value.replace(replacement, "")
        parts.append(value)
        value = ".".join(reversed(parts))
    return value


def get_args(func):
    """Returns the names of the positional arguments for composite model
    inspection."""
    try:
        return list(inspect.signature(func).parameters.keys())[1:]

    except AttributeError:  # pragma: nocover
        return inspect.getargspec(func).args[1:]  # pylint:disable=deprecated-pragma pragma:nocover


def setdefaultattr(obj, name, value):
    """setdefault for object attributes."""
    try:
        return getattr(obj, name)

    except AttributeError:
        setattr(obj, name, value)
    return getattr(obj, name)


def make_args(*args, **kwargs):
    """Useful for setting table args and mapper args on models and other
    things."""
    return tuple(args) + (kwargs,)


def lower(value):
    """Convert value to lowercase if possible.

    For example::

        >>> print(lower('HELLO'))
        hello
        >>> print(lower(5))
        5
    """
    try:
        return value.lower()
    except AttributeError:
        return value



"""Django REST Framework like router for viewsets."""
import itertools
from collections import namedtuple

from django.core.exceptions import ImproperlyConfigured

from .db import meta


try:
    from django.conf.urls import url as re_path
except ImportError:  # pragma: no cover
    from django.urls import re_path


Route = namedtuple("Route", ["url", "mapping", "name", "detail", "initkwargs"])
DynamicRoute = namedtuple("DynamicRoute", ["url", "name", "detail", "initkwargs"])


def escape_curly_brackets(url_path):
    """Double brackets in regex of url_path for escape string formatting."""
    if ("{" and "}") in url_path:
        url_path = url_path.replace("{", "{{").replace("}", "}}")
    return url_path


def action(methods=None, detail=None, url_path=None, url_name=None, **kwargs):
    """Mark a ViewSet method as a routable action.

    Set the `detail` boolean to determine if this action should apply to
    instance/detail requests or collection/list requests.
    """
    methods = ["get"] if (methods is None) else methods
    methods = [method.lower() for method in methods]

    assert detail is not None, "@action() missing required argument: 'detail'"

    def decorator(func):
        func.bind_to_methods = methods
        func.detail = detail
        func.url_path = url_path or func.__name__
        func.url_name = url_name or func.__name__.replace("_", "-")
        func.kwargs = kwargs
        return func

    return decorator


class BaseRouter:
    """Base router."""

    def __init__(self):
        self.registry = []

    def register(self, prefix, viewset, base_name=None):
        """Registers a viewset for route generation."""
        if base_name is None:
            base_name = self.get_default_base_name(viewset)
        self.registry.append((prefix, viewset, base_name))

    def get_default_base_name(self, viewset):
        """If `base_name` is not specified, attempt to automatically determine
        it from the viewset."""
        raise NotImplementedError("get_default_base_name must be overridden")

    def get_urls(self):
        """Return a list of URL patterns, given the registered viewsets."""
        raise NotImplementedError("get_urls must be overridden")

    @property
    def urls(self):
        """URL's routed."""
        if not hasattr(self, "_urls"):
            self._urls = self.get_urls()
        return self._urls


class SimpleRouter(BaseRouter):
    """Generates url patterns that map requests to a viewset's action
    functions.

    It will map the following operations to following actions on the viewset:

    ====== ======================== =============== =================
    Method Path                     Action          Route Name
    ====== ======================== =============== =================
    GET    /<resource>/             list            <resource>-list
    POST   /<resource>/             create          <resource>-list
    GET    /<resource>/new/         new             <resource>-new
    GET    /<resource>/<pk>/        retrieve        <resource>-detail
    POST   /<resource>/<pk>/        update          <resource>-detail
    PUT    /<resource>/<pk>/        update          <resource>-detail
    PATCH  /<resource>/<pk>/        update          <resource>-detail
    DELETE /<resource>/<pk>/        destroy         <resource>-detail
    GET    /<resource>/<pk>/edit/   edit            <resource>-edit
    GET    /<resource>/<pk>/delete/ confirm_destoy  <resource>-delete
    POST   /<resource>/<pk>/delete/ destroy         <resource>-delete
    ====== ======================== =============== =================
    """

    routes = [
        # List route.
        Route(
            url=r"^{prefix}{trailing_slash}$",
            mapping={"get": "list", "post": "create"},
            name="{basename}-list",
            detail=False,
            initkwargs={"suffix": "List"},
        ),
        Route(
            url=r"^{prefix}/new{trailing_slash}$",
            mapping={"get": "new"},
            name="{basename}-new",
            detail=False,
            initkwargs={"suffix": "New"},
        ),
        DynamicRoute(
            url=r"^{prefix}/{url_path}{trailing_slash}$", name="{basename}-{url_name}", detail=False, initkwargs={}
        ),
        # Detail route.
        Route(
            url=r"^{prefix}/{lookup}{trailing_slash}$",
            mapping={"get": "retrieve", "post": "update", "put": "update", "patch": "update", "delete": "destroy"},
            name="{basename}-detail",
            detail=True,
            initkwargs={"suffix": "Instance"},
        ),
        Route(
            url=r"^{prefix}/{lookup}/edit{trailing_slash}$",
            mapping={"get": "edit"},
            name="{basename}-edit",
            detail=True,
            initkwargs={"suffix": "Instance"},
        ),
        Route(
            url=r"^{prefix}/{lookup}/delete{trailing_slash}$",
            mapping={"get": "confirm_destroy", "post": "destroy"},
            name="{basename}-destroy",
            detail=True,
            initkwargs={"suffix": "Instance"},
        ),
        DynamicRoute(
            url=r"^{prefix}/{lookup}/{url_path}{trailing_slash}$",
            name="{basename}-{url_name}",
            detail=True,
            initkwargs={},
        ),
    ]

    def __init__(self, trailing_slash=True):
        self.trailing_slash = "/" if trailing_slash else ""
        super().__init__()

    def get_default_base_name(self, viewset):
        """If `base_name` is not specified, attempt to automatically determine
        it from the viewset."""
        model = getattr(viewset, "get_model", lambda: None)()

        assert model is not None, (
            "`base_name` argument not specified, and could not automatically determine the name from the viewset, "
            "as either queryset is is missing or is not a sqlalchemy query, or the serializer_class is not a "
            "sqlalchemy model serializer"
        )

        return model.__name__.lower()

    def get_routes(self, viewset):
        """Augment `self.routes` with any dynamically generated routes.

        Returns a list of the Route namedtuple.
        """
        # converting to list as iterables are good for one pass, known host needs to be checked again and again for
        # different functions.
        known_actions = itertools.chain(*[route.mapping.values() for route in self.routes if isinstance(route, Route)])
        extra_actions = viewset.get_extra_actions()

        # checking action names against the known actions list
        not_allowed = [action.__name__ for action in extra_actions if action.__name__ in known_actions]
        if not_allowed:
            msg = "Cannot use the @action decorator on the following methods, as they are existing routes: %s"
            raise ImproperlyConfigured(msg % ", ".join(not_allowed))

        # partition detail and list actions
        detail_actions = [action for action in extra_actions if action.detail]
        list_actions = [action for action in extra_actions if not action.detail]

        routes = []
        for route in self.routes:
            if isinstance(route, DynamicRoute) and route.detail:
                routes += [self._get_dynamic_route(route, action) for action in detail_actions]
            elif isinstance(route, DynamicRoute):
                routes += [self._get_dynamic_route(route, action) for action in list_actions]
            else:
                routes.append(route)

        return routes

    def _get_dynamic_route(self, route, action):
        initkwargs = route.initkwargs.copy()
        initkwargs.update(action.kwargs)

        url_path = escape_curly_brackets(action.url_path)

        return Route(
            url=route.url.replace("{url_path}", url_path),
            mapping={http_method: action.__name__ for http_method in action.bind_to_methods},
            name=route.name.replace("{url_name}", action.url_name),
            detail=route.detail,
            initkwargs=initkwargs,
        )

    def get_method_map(self, viewset, method_map):
        """Given a viewset, and a mapping of http methods to actions, return a
        new mapping which only includes any mappings that are actually
        implemented by the viewset."""
        return {method: action for method, action in method_map.items() if hasattr(viewset, action)}

    def get_lookup_regex(self, viewset, lookup_prefix=""):
        """Given a viewset, return the portion of URL regex that is used to
        match against a single instance.

        Note that lookup_prefix is not used directly inside REST
        rest_framework itself, but is required in order to nicely
        support nested router implementations, such as drf-nested-
        routers.


        https://github.com/alanjds/drf-nested-routers
        """
        lookup_url_regex = getattr(viewset, "lookup_url_regex", None)
        if lookup_url_regex:
            return lookup_url_regex

        base_regex = "(?P<{lookup_prefix}{lookup_url_kwarg}>{lookup_value})"

        model = getattr(viewset, "get_model", lambda: None)()
        if model:
            info = meta.model_info(model)

            regexes = [
                base_regex.format(
                    lookup_prefix=lookup_prefix,
                    lookup_url_kwarg=key,
                    lookup_value="[^/.]+",
                )
                for key, _ in info.primary_keys.items()
            ]

            return "/".join(regexes)

        lookup_field = getattr(viewset, "lookup_field", "pk")
        lookup_url_kwarg = getattr(viewset, "lookup_url_kwarg", None) or lookup_field
        lookup_value = getattr(viewset, "lookup_value_regex", "[^/.]+")
        return base_regex.format(
            lookup_prefix=lookup_prefix, lookup_url_kwarg=lookup_url_kwarg, lookup_value=lookup_value
        )

    def get_urls(self):
        """Use the registered viewsets to generate a list of URL patterns."""
        ret = []

        for prefix, viewset, basename in self.registry:
            lookup = self.get_lookup_regex(viewset)
            routes = self.get_routes(viewset)

            for route in routes:
                # Only actions which actually exist on the viewset will be bound
                mapping = self.get_method_map(viewset, route.mapping)
                if not mapping:
                    continue

                # Build the url pattern
                regex = route.url.format(prefix=prefix, lookup=lookup, trailing_slash=self.trailing_slash)

                # If there is no prefix, the first part of the url is probably
                #   controlled by project's urls.py and the router is in an app,
                #   so a slash in the beginning will (A) cause Django to give
                #   warnings and (B) generate URLS that will require using '//'.
                if not prefix and regex[:2] == "^/":
                    regex = f"^{regex[2:]}"

                initkwargs = route.initkwargs.copy()
                initkwargs.update({"basename": basename, "detail": route.detail})

                view = viewset.as_view(mapping, **initkwargs)
                name = route.name.format(basename=basename)
                ret.append(re_path(regex, view, name=name))

        return ret



import six

from django.test import TestCase
from django_sorcery.management.commands.sorcery_history import Command

from .base import MigrationMixin


MIGRATION = '''"""zero

Revision ID: 000000000000
Revises:
Create Date: 2018-07-24 01:28:40.893136

"""


# revision identifiers, used by Alembic.
revision = '000000000000'
down_revision = None
branch_labels = ('zero', )
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
'''


class TestHistory(MigrationMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.write_migration(MIGRATION, "000000000000_zero.py")

    def tearDown(self):
        super().tearDown()
        self.delete_migration("000000000000_zero.py")

    def test(self):
        out = six.StringIO()

        cmd = Command(stdout=out)

        cmd.run_from_argv(["./manage.py sorcery", "history", "--no-color"])

        out.seek(0)
        self.assertEqual(
            out.readlines(),
            [
                "Migrations for tests_testapp on database test\n",
                "<base> -> 000000000000 (zero) (head), zero\n",
                "Migrations for tests_otherapp on database test\n",
            ],
        )

    def test_with_range_noapp(self):
        out = six.StringIO()

        cmd = Command(stdout=out)

        with self.assertRaises(SystemExit):
            cmd.run_from_argv(["./manage.py sorcery", "history", "-r", "base:head", "--no-color"])

    def test_with_range(self):
        out = six.StringIO()

        cmd = Command(stdout=out)

        cmd.run_from_argv(["./manage.py sorcery", "history", "tests_testapp", "-r", "base:head", "--no-color"])

        out.seek(0)
        self.assertEqual(
            out.readlines(),
            ["Migrations for tests_testapp on database test\n", "<base> -> 000000000000 (zero) (head), zero\n"],
        )

    def test_bad_range(self):
        out = six.StringIO()
        err = six.StringIO()

        cmd = Command(stdout=out, stderr=err)

        with self.assertRaises(SystemExit):
            cmd.run_from_argv(["./manage.py sorcery", "history", "tests_testapp", "-r", "base-head", "--no-color"])

        err.seek(0)
        self.assertEqual(err.readlines(), ["CommandError: History range requires [start]:[end], [start]:, or :[end]\n"])

    def test_not_alembic_app(self):
        out = six.StringIO()
        err = six.StringIO()

        cmd = Command(stdout=out, stderr=err)

        with self.assertRaises(SystemExit):
            cmd.run_from_argv(["./manage.py sorcery", "history", "foo", "--no-color"])

        err.seek(0)
        self.assertEqual(err.readlines(), ["CommandError: App 'foo' could not be found. Is it in INSTALLED_APPS?\n"])



"""Namespaced Django command."""
import inspect
import os

import sqlalchemy as sa
from django.core.management.base import BaseCommand, CommandParser


class NamespacedCommand(BaseCommand):
    """Namespaced django command implementation."""

    @property
    def commands(self):
        """Returns the subcommands in the namespace."""
        if not hasattr(self, "_commands"):
            self._commands = {}
            for _ in reversed(self.__class__.mro()):
                self._commands.update(
                    {
                        name: cmd_cls
                        for name, cmd_cls in vars(self.__class__).items()
                        if inspect.isclass(cmd_cls) and hasattr(cmd_cls, "mro") and BaseCommand in cmd_cls.mro()
                    }
                )

        return self._commands

    def run_command_from_argv(self, command, argv):
        """Runs the subcommand with namespace adjusted argv."""
        command.style = self.style

        cmd_args = argv[:]
        cmd_args[0] = " ".join([cmd_args[0], cmd_args.pop(1)])
        command.run_from_argv(cmd_args)

    def run_from_argv(self, argv):
        self._called_from_command_line = True

        if len(argv) > 2 and not argv[2].startswith("-") and argv[2] in self.commands:
            command = self.commands.get(argv[2])(stdout=self.stdout, stderr=self.stderr)
            self.run_command_from_argv(command, argv)
        else:
            self.print_help(argv[0], argv[1])

    def create_parser(self, prog_name, subcommand):
        # for django<2.1 compat, filter kwargs
        args = sa.util.get_cls_kwargs(CommandParser)
        kwargs = {
            "cmd": None,
            "prog": f"{os.path.basename(prog_name)} {subcommand}",
            "description": self.help or None,
        }
        parser = CommandParser(**{k: v for k, v in kwargs.items() if k in args})

        for name, command_cls in self.commands.items():
            parser.add_argument(name, help=command_cls.help)

        return parser

    def print_help(self, prog_name, subcommand):
        parser = self.create_parser(prog_name, subcommand)
        parser.print_help(self.stdout)



import unittest

import sqlalchemy as sa
from django.conf import settings
from django_sorcery.db import databases
from django_sorcery.db.sqlalchemy import SQLAlchemy
from django_sorcery.db.utils import dbdict, index_foreign_keys

from ..base import mock
from ..models_multidb import Bar, Foo, default_db, other_db


class TestDbDict(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.settings = mock.patch.dict(settings.SQLALCHEMY_CONNECTIONS, {"minimal": {"DIALECT": "sqlite"}})
        self.settings.start()
        self.addCleanup(self.settings.stop)

    def test_get(self):
        databases = dbdict()

        db = databases.get("minimal")
        self.assertEqual(str(db.url), "sqlite://")

    def test_get_again(self):
        databases = dbdict()

        db1 = databases.get("minimal")

        db2 = databases.get("minimal")
        self.assertEqual(db1, db2)

    def test_update(self):
        databases = dbdict()

        databases.update({"one": SQLAlchemy("sqlite://")}, two=SQLAlchemy("sqlite://"))

    def test_setitem(self):
        databases = dbdict()
        databases["one"] = SQLAlchemy("sqlite://")

    def test_setitem_again(self):
        databases = dbdict()
        databases["one"] = SQLAlchemy("sqlite://")

        with self.assertRaises(RuntimeError):
            databases["one"] = SQLAlchemy("sqlite://")

    def test_setitem_again_same_db(self):
        databases = dbdict()

        db = SQLAlchemy("sqlite://")
        databases["one"] = db

        with self.assertRaises(RuntimeError):
            databases["two"] = db

    def test_setitem_dummy(self):
        databases = dbdict()

        with self.assertRaises(RuntimeError):
            databases["one"] = {}


class TestMultiDbAtomic(unittest.TestCase):
    def setUp(self):
        super().setUp()
        Foo.objects.delete()
        Bar.objects.delete()
        databases.commit()

    def tearDown(self):
        super().tearDown()
        Foo.objects.delete()
        Bar.objects.delete()
        databases.commit()

    def test_multidb(self):
        with databases.atomic():
            default_db.add(Foo(name="1234"))
            other_db.add(Bar(name="1234"))

        self.assertEqual(Foo.objects.count(), 1)
        self.assertEqual(Bar.objects.count(), 1)
        databases.rollback()

        with self.assertRaises(Exception) as ctx, databases.atomic():
            default_db.add(Foo())
            other_db.add(Bar())

        self.assertIsInstance(ctx.exception, sa.exc.IntegrityError)
        self.assertEqual(Foo.objects.count(), 0)
        self.assertEqual(Bar.objects.count(), 0)
        databases.rollback()

    def test_multidb_flush(self):
        default_db.add(Foo(name="1234"))
        other_db.add(Bar(name="1234"))

        databases.flush()

        self.assertEqual(len(default_db.new) + len(other_db.new), 0)

    def test_multidb_commit(self):
        default_db_session = default_db()
        other_db_session = other_db()

        databases.remove()

        self.assertIsNot(default_db(), default_db_session)
        self.assertIsNot(other_db(), other_db_session)


class TestIndexForeignKeys(unittest.TestCase):
    def setUp(self):
        super().setUp()

        self.metadata = sa.MetaData()
        self.parent = sa.Table(
            "parent",
            self.metadata,
            sa.Column("id", sa.Integer(), primary_key=True),
            sa.Column("tenant_id", sa.Integer(), primary_key=True),
        )

        self.child = sa.Table(
            "child",
            self.metadata,
            sa.Column("id", sa.Integer(), primary_key=True),
            sa.Column("tenant_id", sa.Integer(), primary_key=True),
            sa.Column("parent_id", sa.Integer()),
            sa.ForeignKeyConstraint(["tenant_id", "parent_id"], ["parent.tenant_id", "parent.id"]),
        )

    def test_metadata(self):
        index_foreign_keys(self.metadata)

        self.assertEqual(len(self.metadata.tables["child"].indexes), 1)
        ix = next(iter(self.metadata.tables["child"].indexes))
        self.assertEqual(list(ix.columns), [self.child.c.tenant_id, self.child.c.parent_id])

    def test_table(self):
        index_foreign_keys(self.child)

        self.assertEqual(len(self.metadata.tables["child"].indexes), 1)
        ix = next(iter(self.metadata.tables["child"].indexes))
        self.assertEqual(list(ix.columns), [self.child.c.tenant_id, self.child.c.parent_id])



from ..base import TestCase
from ..testapp.views import OwnerViewSet


class TestViewSet(TestCase):
    def test_as_view_without_action_map(self):
        with self.assertRaises(TypeError):
            OwnerViewSet.as_view()

    def test_as_view_with_verb_initkwargs(self):
        with self.assertRaises(TypeError):
            OwnerViewSet.as_view(actions={"get": "list"}, get="test")

    def test_as_view_with_bad_initkwargs(self):
        with self.assertRaises(TypeError):
            OwnerViewSet.as_view(actions={"get": "list"}, foo="test")

    def test_as_view(self):
        view = OwnerViewSet.as_view(actions={"get": "list"})

        self.assertEqual(view.cls, OwnerViewSet)
        self.assertEqual(view.actions, {"get": "list"})
        self.assertEqual(view.initkwargs, {})
        self.assertIsNone(view.suffix)

        request = self.factory.get("/")
        response = view(request)
        self.assertEqual(response.status_code, 200)



from enum import Enum

from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django_sorcery.db import databases
from django_sorcery.db.models import autocoerce, autocoerce_properties
from django_sorcery.db.query import Query
from django_sorcery.validators import (
    ValidateTogetherModelFields,
    ValidateUnique,
)


db = databases.get("test")


COLORS = ["", "red", "green", "blue", "silver", "pink"]


class VehicleType(Enum):
    bus = "Bus"
    car = "Car"


class States(Enum):
    NY = "NY"
    NJ = "NJ"


class OwnerQuery(Query):
    pass


def street_validator(value):
    if len(value) < 3:
        raise ValidationError("Street should be at least 2 characters.")


class Address(db.BaseComposite):
    street = db.CharField(length=300, validators=[street_validator])
    state = db.EnumField(choices=States, constraint_name="states")
    zip = db.CharField(length=15, validators=[RegexValidator(r"^\d+$")])

    validators = [ValidateTogetherModelFields(["street", "state", "zip"])]

    def clean_state(self):
        if isinstance(self.state, str) and self.state != self.state.upper():
            raise ValidationError({"state": "State must be uppercase."})

    def clean_zip(self):
        if self.zip.startswith("0"):
            raise ValidationError("Zip cannot start with 0.")


class Business(db.Model):
    id = db.IntegerField(autoincrement=True, primary_key=True)
    name = db.CharField(length=50)
    employees = db.IntegerField(default=5, nullable=False)

    location = db.CompositeField(Address)
    other_location = db.CompositeField(Address, prefix="foo")

    class Meta:
        db_table = "business"

    def clean(self, **kwargs):
        if self.other_location and not self.location:
            raise ValidationError({"location": "Primary location is required when other location is provided."})


class Owner(db.Model):
    query_class = OwnerQuery

    id = db.IntegerField(autoincrement=True, primary_key=True)
    first_name = db.CharField(length=50)
    last_name = db.CharField(length=50)

    def clean_first_name(self):
        if self.first_name == "invalid":
            raise ValidationError("Invalid first name")


class Vehicle(db.Model):
    id = db.IntegerField(autoincrement=True, primary_key=True, doc="The primary key")
    name = db.CharField(doc="The name of the vehicle", label="Name", length=50)
    type = db.EnumField(choices=VehicleType, constraint_name="ck_vehicle_type", nullable=False)
    created_at = db.DateTimeField()
    paint = db.EnumField(choices=COLORS, constraint_name="ck_colors")
    is_used = db.BooleanField()
    msrp = db.DecimalField(max_digits=10, decimal_places=2)

    owner = db.ManyToOne(Owner, backref="vehicles")

    @property
    def lower_name(self):
        return self.name.lower()

    @db.validates("name")
    def validate_name(self, key, value):
        if value == "Bad Vehicle":
            raise ValidationError("Name cannot be `Bad Value`")

        return value

    def clean_paint(self):
        if self.paint == "pink":
            raise ValidationError("Can't have a pink car")


class Part(db.Model):
    id = db.IntegerField(autoincrement=True, primary_key=True)
    name = db.CharField(length=50)

    vehicles = db.ManyToMany(Vehicle, backref="parts", table_name="vehicle_parts")


class Option(db.Model):
    id = db.IntegerField(autoincrement=True, primary_key=True)
    name = db.CharField(length=50)

    vehicles = db.ManyToMany(Vehicle, backref=db.backref("options"), table_name="vehicle_options")


class CompositePkModel(db.Model):
    query_class = None

    id = db.IntegerField(primary_key=True)
    pk = db.IntegerField(primary_key=True)
    name = db.CharField(length=50)

    is_active = db.BooleanField()

    active = db.queryproperty(is_active=True)


class DummyEnum(Enum):
    zero = 0
    one = 1
    two = 2


Choices = ["three", "four"]


@autocoerce
class AllKindsOfFields(db.Model):
    pk = db.Column(db.Integer(), primary_key=True)

    # flags
    boolean_notnull = db.Column(db.Boolean(), nullable=False)
    boolean = db.Column(db.Boolean())
    enum = db.Column(db.Enum(DummyEnum, name="dummy_enum"))
    enum_choice = db.Column(db.Enum(*Choices, name="some_choices"))

    # numbers
    bigint = db.Column(db.BIGINT())
    biginteger = db.Column(db.BigInteger())
    decimal = db.Column(db.DECIMAL())
    float = db.Column(db.Float())
    int = db.Column(db.INT())
    integer = db.Column(db.Integer())
    numeric = db.Column(db.Numeric())
    real = db.Column(db.REAL())
    smallint = db.Column(db.SMALLINT())
    smallinteger = db.Column(db.SmallInteger())

    # strings
    char = db.Column(db.CHAR())
    # clob = db.Column(db.CLOB())
    nchar = db.Column(db.NCHAR())
    # nvarchar = db.Column(db.NVARCHAR())
    string = db.Column(db.String())
    text = db.Column(db.Text())
    unicode = db.Column(db.Unicode())
    unicodetext = db.Column(db.UnicodeText())
    varchar = db.Column(db.VARCHAR())

    # dates, times and durations
    date = db.Column(db.Date())
    datetime = db.Column(db.DateTime())
    interval = db.Column(db.Interval())
    time = db.Column(db.Time())
    timestamp = db.Column(db.TIMESTAMP())

    # blobs
    # binary = db.Column(db.Binary())
    # blob = db.Column(db.BLOB())
    largebinary = db.Column(db.LargeBinary())
    # varbinary = db.Column(db.VARBINARY())


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __composite_values__(self):
        return self.x, self.y

    def __repr__(self):
        return "Point(x={!r}, y={!r})".format(self.x, self.y)

    def __eq__(self, other):
        return isinstance(other, Point) and other.x == self.x and other.y == self.y

    def __ne__(self, other):
        return not self.__eq__(other)


class Vertex(db.Model):
    pk = db.IntegerField(autoincrement=True, primary_key=True)

    x1 = db.IntegerField()
    y1 = db.IntegerField()
    x2 = db.IntegerField()
    y2 = db.IntegerField()

    start = db.composite(Point, x1, y1)
    end = db.composite(Point, x2, y2)


class ModelTwo(db.Model):
    pk = db.IntegerField(autoincrement=True, primary_key=True)
    name = db.CharField(length=50)


class ModelOne(db.Model):
    pk = db.IntegerField(autoincrement=True, primary_key=True)
    name = db.CharField(length=50)

    _model_twos = db.OneToMany(ModelTwo, backref="_model_one")


model_three_ones = db.Table(
    "model_three_ones",
    db.Column("model_one_pk", db.Integer(), db.ForeignKey("model_one.pk"), primary_key=True),
    db.Column("model_two_pk", db.Integer(), db.ForeignKey("model_three.pk"), primary_key=True),
)


class ModelThree(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String())

    model_ones = db.ManyToMany(ModelOne, secondary=model_three_ones)


class ModelFour(db.Model):
    pk = db.IntegerField(autoincrement=True, primary_key=True)
    name = db.CharField(length=50)

    _model_twos = db.OneToMany(ModelTwo, backref=db.backref("_model_four"))


class ModelFullCleanFail(db.Model):
    pk = db.IntegerField(autoincrement=True, primary_key=True)
    name = db.CharField(length=50)

    def clean(self, **kwargs):
        raise ValidationError("bad model")


class ValidateUniqueModel(db.Model):
    pk = db.IntegerField(autoincrement=True, primary_key=True)
    name = db.CharField(length=50)
    foo = db.CharField(length=50)
    bar = db.CharField(length=50)

    validators = [ValidateUnique(db, "name"), ValidateUnique(db, "foo", "bar")]


class ClassicModel:
    pass


class SelectedAutoCoerce(db.Model):
    pk = db.IntegerField(autoincrement=True, primary_key=True)
    foo = db.DecimalField()
    bar = db.FloatField()
    me = db.IntegerField()


autocoerce_properties(SelectedAutoCoerce.foo, SelectedAutoCoerce.bar)


classic_model_table = db.Table(
    "classic_model", db.Column("pk", db.Integer(), autoincrement=True, primary_key=True), db.Column("name", db.String())
)
db.mapper(ClassicModel, classic_model_table)

db.configure_mappers()
db.create_all()



import six

from django.test import TestCase
from django_sorcery.db import databases
from django_sorcery.management.commands.sorcery_downgrade import Downgrade
from django_sorcery.management.commands.sorcery_stamp import Command

from .base import MigrationMixin


M1 = '''"""zero

Revision ID: 000000000000
Revises:
Create Date: 2018-07-24 01:28:40.893136

"""


# revision identifiers, used by Alembic.
revision = '000000000000'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
'''

db = databases.get("test")


class TestStamp(MigrationMixin, TestCase):
    def setUp(self):
        super().setUp()

        self.delete_migration("{}_.py".format("000000000000"))
        self.write_migration(M1, "{}_.py".format("000000000000"))

    def tearDown(self):
        super().tearDown()

        Downgrade().run_from_argv(["./manage.py sorcery", "downgrade", "--no-color"])
        self.delete_migration("{}_.py".format("000000000000"))
        Downgrade().run_from_argv(["./manage.py sorcery", "downgrade", "--no-color"])

    def test(self):
        out = six.StringIO()
        cmd = Command(stdout=out)
        cmd.run_from_argv(["./manage.py sorcery", "stamp", "tests_testapp", "--no-color"])

        revs = db.execute("select * from alembic_version_tests_testapp").fetchall()
        self.assertEqual(revs, [("000000000000",)])

    def test_bad_rev(self):
        out = six.StringIO()
        cmd = Command(stdout=out)

        with self.assertRaises(SystemExit):
            cmd.run_from_argv(["./manage.py sorcery", "stamp", "tests_testapp", "-r", ":000000000000", "--no-color"])



"""sqlalchemy profiling things."""
import logging
import time
from collections import defaultdict, namedtuple
from functools import partial
from threading import local

import sqlalchemy as sa
from django.conf import settings


logger = logging.getLogger(__name__)
STATEMENT_TYPES = {"SELECT": "select", "INSERT INTO": "insert", "UPDATE": "update", "DELETE": "delete"}


Query = namedtuple("Query", ["timestamp", "statement", "parameters", "duration"])


class SQLAlchemyProfiler:
    """A sqlalchemy profiler that hooks into sqlalchemy engine and pool events
    and generate stats.

    Can also capture executed sql statements. Useful for profiling or
    testing sql statements.
    """

    def __init__(self, exclude=None, record_queries=True):
        self.local = local()
        self.exclude = exclude or []
        self.record_queries = record_queries

        self._events = [
            ("before_cursor_execute", sa.engine.Engine, self._before_cursor_execute),
            ("after_cursor_execute", sa.engine.Engine, self._after_cursor_execute),
            ("begin", sa.engine.Engine, partial(self._event_counter, count_event="begin")),
            ("begin_twophase", sa.engine.Engine, partial(self._event_counter, count_event="begin_twophase")),
            ("commit", sa.engine.Engine, partial(self._event_counter, count_event="commit")),
            ("commit_twophase", sa.engine.Engine, partial(self._event_counter, count_event="commit_twophase")),
            ("prepare_twophase", sa.engine.Engine, partial(self._event_counter, count_event="prepare_twophase")),
            ("release_savepoint", sa.engine.Engine, partial(self._event_counter, count_event="prepare_twophase")),
            ("rollback", sa.engine.Engine, partial(self._event_counter, count_event="rollback")),
            ("rollback_savepoint", sa.engine.Engine, partial(self._event_counter, count_event="rollback_savepoint")),
            ("rollback_twophase", sa.engine.Engine, partial(self._event_counter, count_event="rollback_twophase")),
            ("savepoint", sa.engine.Engine, partial(self._event_counter, count_event="savepoint")),
            ("dbapi_error", sa.engine.Engine, partial(self._event_counter, count_event="dbapi_error")),
            ("engine_connect", sa.engine.Engine, partial(self._event_counter, count_event="engine_connect")),
            ("engine_disposed", sa.engine.Engine, partial(self._event_counter, count_event="engine_disposed")),
            ("checkin", sa.pool.Pool, partial(self._event_counter, count_event="pool_checkin")),
            ("checkout", sa.pool.Pool, partial(self._event_counter, count_event="pool_checkout")),
            ("close", sa.pool.Pool, partial(self._event_counter, count_event="pool_close")),
            ("close_detached", sa.pool.Pool, partial(self._event_counter, count_event="pool_close_detached")),
            ("connect", sa.pool.Pool, partial(self._event_counter, count_event="pool_connect")),
            ("detach", sa.pool.Pool, partial(self._event_counter, count_event="pool_detach")),
            ("first_connect", sa.pool.Pool, partial(self._event_counter, count_event="pool_first_connect")),
            ("invalidate", sa.pool.Pool, partial(self._event_counter, count_event="pool_invalidate")),
            ("reset", sa.pool.Pool, partial(self._event_counter, count_event="pool_reset")),
            ("soft_invalidate", sa.pool.Pool, partial(self._event_counter, count_event="pool_soft_invalidate")),
        ]

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def __del__(self):
        # stop all listeners when profiler gets garbage collected
        # py2 does not collect coverage on this
        self.stop()  # pragma: nocover

    def start(self):
        """Starts profiling by wiring up sqlalchemy events."""
        self.clear()
        for ev, target, handler in self._events:
            try:
                if not sa.event.contains(target, ev, handler):
                    sa.event.listen(target, ev, handler)
            except Exception:  # pragma: nocover
                # Gets raised when pool doesnt support the event, so ignore it
                pass  # pragma: nocover

    def stop(self):
        """Stops profiling by detaching wired up sqlalchemy events."""
        for ev, target, handler in self._events:
            try:
                if sa.event.contains(target, ev, handler):
                    sa.event.remove(target, ev, handler)
            except Exception:  # pragma: nocover
                # Gets raised when pool doesnt support the event, so ignore it
                pass  # pragma: nocover

    def clear(self):
        """Clears collected stats."""
        self.local.__dict__.clear()

    @property
    def duration(self):
        """Return total statement execution duration."""
        return self.local.__dict__.setdefault("duration", 0)

    @duration.setter
    def duration(self, value):
        """Sets total statement execution duration."""
        self.local.duration = value

    @property
    def counts(self):
        """Returns a dict of counts per sqlalchemy event operation like
        executed statements, commits, rollbacks, etc.."""
        return self.local.__dict__.setdefault("counts", defaultdict(lambda: 0))

    @property
    def queries(self):
        """Returns executed statements."""
        return self.local.__dict__.setdefault("queries", [])

    @property
    def stats(self):
        """Returns profiling stats."""
        stats = self.counts.copy()
        stats["duration"] = self.duration
        return stats

    def _before_cursor_execute(self, conn, cursor, statement, parameters, context, executemany):
        self.local._profiler_query_start_time = time.time()

    def _after_cursor_execute(self, conn, cursor, statement, parameters, context, executemany):
        end_time = time.time()
        start_time = self.local._profiler_query_start_time
        duration = end_time - start_time

        for e in self.exclude:
            if e in statement:
                return

        params = getattr(context, "compiled_parameters", [])
        if self.record_queries:
            self.queries.append(Query(int(round(time.time() * 1000)), statement, params, duration))

        self.duration += duration
        self.counts["execute"] += 1

        for start, event in STATEMENT_TYPES.items():
            if statement.startswith(start):
                self.counts[event] += 1
                break

    def _event_counter(self, *args, **kwargs):
        count_event = kwargs.get("count_event")
        self.counts[count_event] += 1


class SQLAlchemyProfilingMiddleware:
    """Django middleware that provides sqlalchemy statistics."""

    logger = logger

    def __init__(self, get_response=None):
        self.get_response = get_response
        self.profiler = SQLAlchemyProfiler(record_queries=False)

    @property
    def log_results(self):
        """Determines if stats should be logged or not."""
        return settings.DEBUG

    @property
    def header_results(self):
        """Determines if stats should be returned as headers or not."""
        return settings.DEBUG

    def start(self):
        """Starts profiling and disables restarts."""
        self.profiler.start()
        self.start = lambda: None

    def __call__(self, request):
        self.process_request(request)
        response = self.get_response(request)
        return self.process_response(request, response)

    def process_request(self, request):
        """Starts profiling and resets stats doe the request."""
        self.start()
        self.profiler.clear()

    def process_response(self, request, response):
        """Logs current request stats and also returns stats as headers."""
        try:
            stats = self.profiler.stats
            if stats["duration"] or self.log_results:
                self.log(**{"sa_{}".format(k): v for k, v in stats.items()})
        except Exception:  # pragma: nocover
            # The show must go on...
            pass  # pragma: nocover
        else:
            if self.header_results:
                for k, v in stats.items():
                    response["X-SA-{}".format("".join(i.title() for i in k.split("_")))] = v
        self.profiler.clear()
        return response

    def log(self, **kwargs):
        """Log sqlalchemy stats for current request."""
        self.logger.info("SQLAlchemy profiler %s", " ".join("{}={}".format(k, v) for k, v in kwargs.items()))



from django.apps import AppConfig


class TestAppConfig(AppConfig):
    name = "tests.testapp"
    label = "tests_testapp"
    version_table_schema = "public"

    def ready(self):
        from .models import db

        db.configure_mappers()
        db.create_all()



import sqlalchemy as sa
from django.apps import apps
from django.core.exceptions import FieldDoesNotExist
from django_sorcery.db import meta  # noqa

from ...base import TestCase
from ...otherapp.models import OtherAppInOtherApp
from ...testapp.models import Business, CompositePkModel, Owner, Vehicle


class TestModelMeta(TestCase):
    def test_model_meta(self):
        info = meta.model_info(Owner)

        self.assertEqual(set(info.primary_keys.keys()), {"id"})
        self.assertEqual(info.primary_keys["id"].property, Owner.id.property)

        self.assertEqual(set(info.properties.keys()), {"first_name", "last_name"})
        self.assertEqual(info.properties["first_name"].property, Owner.first_name.property)
        self.assertEqual(info.properties["last_name"].property, Owner.last_name.property)

        self.assertEqual(set(info.relationships.keys()), {"vehicles"})
        self.assertTrue(info.relationships["vehicles"].relationship is Owner.vehicles.property)

        self.assertEqual(set(info.field_names), {"id", "first_name", "last_name", "vehicles"})

        self.assertEqual(
            [attr for attr in dir(info) if not attr.startswith("__")],
            [
                "_init",
                "app_config",
                "app_label",
                "clean_fields",
                "clean_nested_fields",
                "clean_relation_fields",
                "column_properties",
                "composites",
                "concrete_fields",
                "field_names",
                "fields",
                "first_name",
                "full_clean",
                "get_field",
                "get_key",
                "id",
                "identity_key_from_dict",
                "identity_key_from_instance",
                "label",
                "label_lower",
                "last_name",
                "local_fields",
                "mapper",
                "model",
                "model_class",
                "model_name",
                "object_name",
                "opts",
                "ordering",
                "primary_keys",
                "primary_keys_from_dict",
                "primary_keys_from_instance",
                "private_fields",
                "properties",
                "relationships",
                "run_validators",
                "sa_state",
                "unique_together",
                "vehicles",
                "verbose_name",
                "verbose_name_plural",
            ],
        )
        self.assertEqual(info.id, info.primary_keys["id"])
        self.assertEqual(info.first_name, info.properties["first_name"])
        self.assertEqual(info.vehicles, info.relationships["vehicles"])
        with self.assertRaises(AttributeError):
            info.abc

        info = meta.model_info(Business)
        self.assertEqual(info.location, info.composites["location"])
        self.assertEqual(info.other_location, info.composites["other_location"])
        self.assertNotEqual(info.location, info.other_location)

        self.assertEqual(info.label, "tests_testapp.Business")
        self.assertEqual(info.label_lower, "tests_testapp.business")

        info = meta.model_info(Vehicle)
        self.assertEqual(info.private_fields, ())

        self.assertEqual(
            {f.name for f in info.local_fields},
            {"id", "name", "type", "created_at", "paint", "is_used", "msrp", "owner", "parts", "options"},
        )
        for f in info.local_fields:
            self.assertEqual(f, info.get_field(f.name))

    def test_app(self):
        info = meta.model_info(Owner)
        self.assertIs(info.app_config, apps.get_containing_app_config(Owner.__module__))

        info = meta.model_info(OtherAppInOtherApp)
        self.assertIs(info.app_config, apps.get_containing_app_config(Owner.__module__))

    def test_reprs(self):
        owner_info = meta.model_info(Owner)
        self.assertEqual(
            repr(owner_info),
            "\n".join(
                [
                    "<model_info(Owner)>",
                    "    <integer_column_info(Owner.id) pk>",
                    "    <string_column_info(Owner.first_name)>",
                    "    <string_column_info(Owner.last_name)>",
                    "    <relation_info(Owner.vehicles)>",
                ]
            ),
        )

        vehicle_info = meta.model_info(Vehicle)
        self.assertEqual(
            repr(vehicle_info),
            "\n".join(
                [
                    "<model_info(Vehicle)>",
                    "    <integer_column_info(Vehicle.id) pk>",
                    "    <integer_column_info(Vehicle._owner_id)>",
                    "    <datetime_column_info(Vehicle.created_at)>",
                    "    <boolean_column_info(Vehicle.is_used)>",
                    "    <numeric_column_info(Vehicle.msrp)>",
                    "    <string_column_info(Vehicle.name)>",
                    "    <choice_column_info(Vehicle.paint)>",
                    "    <enum_column_info(Vehicle.type)>",
                    "    <relation_info(Vehicle.options)>",
                    "    <relation_info(Vehicle.owner)>",
                    "    <relation_info(Vehicle.parts)>",
                ]
            ),
        )
        business_info = meta.model_info(Business)
        self.assertEqual(
            repr(business_info),
            "\n".join(
                [
                    "<model_info(Business)>",
                    "    <integer_column_info(Business.id) pk>",
                    "    <integer_column_info(Business.employees)>",
                    "    <string_column_info(Business.name)>",
                    "    <composite_info(Address, Business.location)>",
                    "        <enum_column_info(Address.state)>",
                    "        <string_column_info(Address.street)>",
                    "        <string_column_info(Address.zip)>",
                    "    <composite_info(Address, Business.other_location)>",
                    "        <enum_column_info(Address.state)>",
                    "        <string_column_info(Address.street)>",
                    "        <string_column_info(Address.zip)>",
                ]
            ),
        )

    def test_inspect(self):
        info = meta.model_info(Owner)
        self.assertIsInstance(info.sa_state(Owner()), sa.orm.state.InstanceState)

    def test_get_field(self):
        info = meta.model_info(Owner)
        self.assertIs(info.get_field("first_name"), info.properties.get("first_name"))

        with self.assertRaises(FieldDoesNotExist):
            info.get_field("zzzzzz")

    def test_column_properties(self):
        info = meta.model_info(Owner)

        self.assertListEqual(
            list(info.column_properties),
            [("id", info.id), ("first_name", info.first_name), ("last_name", info.last_name)],
        )

    def test_model_meta_with_mapper(self):
        mapper = Vehicle.owner.property.parent
        self.assertEqual(meta.model_info(mapper), meta.model_info(Vehicle))

    def test_primary_keys_from_dict(self):
        info = meta.model_info(Owner)

        key = info.primary_keys_from_dict({"id": 1})
        self.assertEqual(key, 1)

        info = meta.model_info(CompositePkModel)
        key = info.primary_keys_from_dict({"id": 1, "pk": 2})
        self.assertEqual(key, (1, 2))

    def test_primary_keys_from_instance(self):
        info = meta.model_info(Owner)
        instance = Owner(id=1)

        self.assertIsNone(info.primary_keys_from_instance(None))

        key = info.primary_keys_from_instance(instance)
        self.assertEqual(key, 1)

        info = meta.model_info(CompositePkModel)
        instance = CompositePkModel(id=1, pk=2)

        key = info.primary_keys_from_instance(instance)
        self.assertDictEqual(key, {"id": 1, "pk": 2})

    def test_get_key(self):
        info = meta.model_info(Owner)
        owner = Owner(id=1)

        key = info.get_key(owner)
        self.assertEqual(key, (1,))
        self.assertIsNone(info.get_key(Owner()))

    def test_identity_key_from_instance(self):
        info = meta.model_info(Owner)
        owner = Owner(id=1)

        key = info.identity_key_from_instance(owner)
        self.assertEqual(key, meta.Identity(Owner, (1,)))
        self.assertIsNone(info.identity_key_from_instance(Owner()))

    def test_identity_key_from_dict(self):
        info = meta.model_info(Owner)

        key = info.identity_key_from_dict({"id": 1})
        self.assertEqual(key, meta.Identity(Owner, (1,)))
        self.assertIsNone(info.identity_key_from_dict({}))



"""Heads command."""

from ..alembic import AlembicCommand


class ShowHeads(AlembicCommand):
    """Display alembic revision heads."""

    help = "Display revision heads"

    def add_arguments(self, parser):
        parser.add_argument("app_label", nargs="?", help="App label of application to limit the output to.")

    def handle(self, app_label=None, verbosity=0, **kwargs):
        verbosity = bool(verbosity - 1)
        appconfigs = [self.lookup_app(app_label)] if app_label is not None else self.sorcery_apps.values()

        for appconfig in appconfigs:
            self.stdout.write(
                self.style.SUCCESS("Heads for {} on database {}".format(appconfig.name, appconfig.db.alias))
            )
            for rev in appconfig.script.get_revisions("heads"):
                if verbosity:
                    self.stdout.write(
                        "".join(
                            [
                                "[",
                                appconfig.name,
                                "]\n",
                                rev.cmd_format(verbosity, include_branches=True, tree_indicators=False),
                            ]
                        )
                    )
                else:
                    self.stdout.write(
                        "".join(
                            [
                                rev.cmd_format(verbosity, include_branches=True, tree_indicators=False),
                                " <",
                                appconfig.name,
                                "> ",
                            ]
                        )
                    )


Command = ShowHeads



"""Stamp command."""
from functools import partial

import alembic
from django.core.management import CommandError

from ..alembic import AlembicCommand


class Stamp(AlembicCommand):
    """Stamp the revision table with migration revisions, doesn't run any
    migrations."""

    help = "Stamp the revision table with migration revisions, doesn't run any migrations"

    def add_arguments(self, parser):
        parser.add_argument("args", metavar="app_label", nargs=1, help="Specify the app label to stamp for.")
        parser.add_argument(
            "-r",
            "--revision",
            default="heads",
            help="Database state will be brought to the state after that "
            'migration. Use the name "base" to unapply all migrations.',
        )

    def handle(self, app_label=None, revision=None, **kwargs):
        appconfig = self.lookup_app(app_label)

        if ":" in revision:
            raise CommandError("Range revision is not allowed")

        self.stdout.write(
            self.style.SUCCESS("Stamping revision for {} on database {}".format(appconfig.name, appconfig.db.alias))
        )

        with alembic.context.EnvironmentContext(
            appconfig.config,
            appconfig.script,
            fn=partial(self.stamp, appconfig=appconfig, revision=revision),
            destination_rev=revision,
        ) as context:
            self.run_env(context, appconfig)

    def stamp(self, rev, context, appconfig, revision):
        """Stamp the alembic revision."""
        return appconfig.script._stamp_revs(revision, rev)


Command = Stamp



import os
from contextlib import suppress
from importlib import import_module

import sqlalchemy as sa
from django.conf import settings
from django.utils.encoding import force_str
from django.utils.module_loading import import_string


DIALECT_MAP = {
    "django.db.backends.sqlite3": "sqlite",
    "django.db.backends.postgresql": "postgresql",
    "django.db.backends.mysql": "mysql",
    "django.db.backends.oracle": "oracle",
    "sqlserver_ado": "mssql",
}
DIALECT_MAP_TO_DJANGO = {v: k for k, v in DIALECT_MAP.items()}


def boolean(x):
    return str(x) in {"True", "1"}


def integer(x):
    return int(x)


def string(x):
    return force_str(x)


def string_list(x):
    return force_str(x).split(",")


def importable(x):
    try:
        return import_string(x)
    except ImportError:
        return import_module(x)


def importable_list(x):
    return [importable(i) for i in force_str(x).split(",")]


def importable_list_tuples(x):
    return [(importable(i), j) for i, j in [k.split(":") for k in force_str(x).split(",")]]


ENGINE_OPTIONS_NORMALIZATION = {
    "case_sensitive": boolean,
    # connect_args - direct querystring params in url
    "convert_unicode": boolean,
    "creator": importable,
    "echo": boolean,
    "echo_pool": boolean,
    "empty_in_strategy": string,
    "encoding": string,
    # execution_options - signal should be used instead
    "implicit_returning": boolean,
    "isolation_level": string,
    "label_length": integer,
    "listeners": importable_list,
    "logging_name": string,
    "max_overflow": integer,
    "module": importable,
    "paramstyle": string,
    "plugins": string_list,
    "pool": lambda x: importable(x)(),
    "pool_events": importable_list_tuples,
    "pool_logging_name": string,
    "pool_pre_ping": boolean,
    "pool_recycle": integer,
    "pool_reset_on_return": string,
    "pool_size": integer,
    "pool_threadlocal": boolean,
    "pool_timeout": integer,
    "pool_use_lifo": boolean,
    "poolclass": importable,
    "strategy": string,
}


def get_settings(alias):
    """Returns database settings from either ``SQLALCHEMY_CONNECTIONS`` setting
    or ``DATABASES`` setting."""

    if hasattr(settings, "SQLALCHEMY_CONNECTIONS") and alias in settings.SQLALCHEMY_CONNECTIONS:
        return settings.SQLALCHEMY_CONNECTIONS[alias]

    return settings.DATABASES[alias]


def _asdict(url):
    return {
        "drivername": url.drivername,
        "username": url.username,
        "password": getattr(url, "password_original", None) or getattr(url, "password", None),
        "host": url.host,
        "port": url.port,
        "database": url.database,
        "query": dict(url.query.items()),
    }


def _options_from_url(url, base_options):
    options = base_options.copy()
    url_kwargs = _asdict(url)
    query_options = url_kwargs["query"]
    options.update(
        {
            "engine_options": {
                k.replace("engine_", ""): ENGINE_OPTIONS_NORMALIZATION.get(k.replace("engine_", ""), lambda x: x)(
                    query_options.pop(k)
                )
                for k in list(query_options)
                if k.startswith("engine_")
            }
        }
    )
    return sa.engine.url.URL(**url_kwargs), options


def make_url(alias_or_url):
    """
    Generates a URL either from a url string, environment variable, SQLALCHEMY_CONNECTIONS or DATABASES settings
    ---------------------------------------------------------------------
    alias_or_url: str
        name of the alias or url as string
    """
    settings_kwargs = {}
    with suppress(KeyError):
        _, settings_kwargs = make_url_from_settings(alias_or_url)

    try:
        url = sa.engine.url.make_url(alias_or_url)

    except sa.exc.ArgumentError:
        pass

    else:
        return _options_from_url(url, settings_kwargs)

    alias = alias_or_url

    url = sa.engine.url.make_url(os.environ.get(f"{alias.upper()}_URL", None))
    if url:
        return _options_from_url(url, settings_kwargs)

    return make_url_from_settings(alias)


def make_url_from_settings(alias):
    """
    Generates a URL using the alias in settings
    -------------------------------------------
    alias: str
        name of the alias

    Overall settings are very similar with django database settings with a few extra keys.

    ``USER`` - database user

    ``PASSWORD`` - database user password

    ``HOST`` - database host

    ``NAME`` - database name

    ``PORT`` - database name

    ``DIALECT`` - dialect to be used in url, if not provided, will use the ``DIALECT_MAP`` to figure out a dialect to
    be used in sqlalchemy url

    ``DRIVER`` - If provided, will be used as the driver in sqlalchemy url

    ``SQLALCHEMY`` - If provided, a custom :py:class:`..sqlalchemy.SQLAlchemy` class to be used

    ``QUERY`` - querystring arguments for sqlalchemy url

    ``ALCHEMY_OPTIONS`` - Optional arguments to be used to initialize the :py:class:`..sqlalchemy.SQLAlchemy` instance

        * ``session_class`` - a custom session class to be used
        * ``registry_class`` - a custom registy class to be used for scoping
        * ``model_class`` - a custom base model class to be used for declarative base models.
        * ``metadata_class`` - a custom metadata class used in delclarative models.
        * ``metadata_options`` - custom options to use in metadata creation such as specifying naming conventions.
        * ``engine_options`` - arguments for sqlalchemy ``create_engine``
        * ``session_options`` - arguments for sqlalchemy ``sessionmaker``

    Other options are ignored.
    """
    data = get_settings(alias)

    if "DIALECT" not in data:
        data["DIALECT"] = DIALECT_MAP.get(data["ENGINE"]) or data["ENGINE"].split(".")[-1]

    names = [data["DIALECT"].lower()]

    if "DRIVER" in data:
        names.append(data["DRIVER"].lower())

    drivername = "+".join(names)

    kwargs = {
        "username": data.get("USER") or None,
        "password": data.get("PASSWORD") or None,
        "host": data.get("HOST") or None,
        "database": data.get("NAME") or None,
        "query": data.get("QUERY", {}),
    }

    with suppress(Exception):
        kwargs["port"] = int(data.get("PORT"))

    url = sa.engine.url.URL(drivername, **kwargs)
    return url, data.get("ALCHEMY_OPTIONS", {})



"""Some Django like shortcuts that support sqlalchemy models."""
from contextlib import suppress

from django.core.exceptions import ImproperlyConfigured
from django.http import Http404
from sqlalchemy.exc import InvalidRequestError


def _get_query(klass):
    query = getattr(klass, "query", None) or getattr(klass, "objects", None)

    if query:
        return query

    with suppress(AttributeError, InvalidRequestError):
        klass._only_full_mapper_zero("get")
        return klass

    raise ImproperlyConfigured("Couldn't figure out the query for model {cls}".format(cls=klass.__name__))


def get_object_or_404(klass, *args, **kwargs):
    """Use session.get() to return an object, or raise a Http404 exception if
    the object does not exist.

    klass may be a Model, or Query object. All other passed arguments
    and keyword arguments are used in the get() query.
    """
    query = _get_query(klass)

    instance = query.get(*args, **kwargs)

    if instance is None:
        raise Http404(f"No {query} matches the given query.")

    return instance


def get_list_or_404(klass, *args, **kwargs):
    """Use filter() to return a list of objects, or raise a Http404 exception
    if the count is 0.

    klass may be a Model or Query object. All other passed arguments
    used in filter() and keyword arguments are used in filter_by().
    """
    query = _get_query(klass)

    if args:
        query = query.filter(*args)

    if kwargs:
        query = query.filter_by(**kwargs)

    if not query.count():
        raise Http404(f"No {query} matches the given query.")

    return query



"""Sorcery command namespace."""

from .. import NamespacedCommand
from .sorcery_createall import CreateAll
from .sorcery_current import Current
from .sorcery_downgrade import Downgrade
from .sorcery_dropall import DropAll
from .sorcery_heads import ShowHeads
from .sorcery_history import History
from .sorcery_revision import Revision
from .sorcery_stamp import Stamp
from .sorcery_upgrade import Upgrade


class Command(NamespacedCommand):
    """Namespaced commands for sorcery."""

    help = "django-sorcery management commands"

    createall = CreateAll
    dropall = DropAll
    history = History
    revision = Revision
    heads = ShowHeads
    upgrade = Upgrade
    downgrade = Downgrade
    current = Current
    stamp = Stamp

    class Meta:
        namespace = "sorcery"



"""Exceptions."""

from django.core.exceptions import ValidationError


class NestedValidationError(ValidationError):
    """Django Validation error except which allows nested errors.

    Useful for validating composite objects.

    For example::

        raise NestedValidationError({
            "field": ["error"],
            "composite": {
                "field": ["error"],
            }
        })
    """

    def __init__(self, message, code=None, params=None):
        if isinstance(message, dict):
            self.error_dict = {}

            for field, messages in message.items():
                _messages = messages
                if not isinstance(messages, ValidationError):
                    _messages = NestedValidationError(messages)
                try:
                    self.error_dict[field] = _messages.error_list
                except AttributeError:
                    self.error_dict[field] = _messages.error_dict

        elif isinstance(message, list):
            self.error_list = []

            for m in message:
                _m = m
                if not isinstance(m, ValidationError):
                    _m = NestedValidationError(m)
                try:
                    self.error_list.extend(_m.error_list)
                except AttributeError:
                    self.error_list.append(_m)

        elif isinstance(message, NestedValidationError):
            self.__dict__.update(vars(message))

        else:
            super().__init__(message, code, params)

    def update_error_dict(self, error_dict):
        if not hasattr(self, "error_dict"):
            return super().update_error_dict(error_dict)

        for field, errors in self.error_dict.items():
            holder = error_dict.setdefault(field, errors.__class__())
            (getattr(holder, "update", None) or holder.extend)(errors)
        return error_dict

    def __iter__(self):
        if hasattr(self, "code"):
            yield from super().__iter__()

        elif hasattr(self, "error_dict"):
            for field, errors in self.error_dict.items():
                errors = NestedValidationError(errors)
                if hasattr(errors, "error_dict"):
                    yield field, dict(errors)
                else:
                    yield field, list(errors)

        else:
            for errors in self.error_list:
                errors = NestedValidationError(errors)
                if hasattr(errors, "error_dict"):
                    yield dict(errors)
                else:
                    yield from errors

    def __repr__(self):
        return "Nested" + super().__repr__()



from django.test import TestCase
from django_sorcery.db import signals
from django_sorcery.db.alembic.signals import (
    include_object,
    process_revision_directives,
)


class TestIncludeObject(TestCase):
    def tearDown(self):
        super().tearDown()
        signals.alembic_include_object._clear_state()

    def test_default(self):
        self.assertTrue(include_object(None, None, None, None, None))

    def test_true(self):
        @signals.alembic_include_object.connect
        def o(*args, **kwargs):
            return True

        self.assertTrue(include_object(None, None, None, None, None))

    def test_false(self):
        @signals.alembic_include_object.connect
        def o(*args, **kwargs):
            return False

        self.assertFalse(include_object(None, None, None, None, None))


class TestProcessRevisionDirectives(TestCase):
    def setUp(self):
        super().tearDown()
        signals.alembic_process_revision_directives.connect(self._handler)

    def tearDown(self):
        super().tearDown()
        signals.alembic_process_revision_directives.disconnect(self._handler)

    def _handler(self, context, revision, directives):
        self.context = context
        self.revision = revision
        self.directives = directives

    def test_process_revision_directives(self):
        process_revision_directives({}, "abc", ("1234",))

        self.assertEqual(self.context, {})
        self.assertEqual(self.revision, "abc")
        self.assertEqual(self.directives, ("1234",))



from .base import GenericViewSet, ModelViewSet, ReadOnlyModelViewSet
from .mixins import (
    CreateModelMixin,
    DeleteModelMixin,
    ListModelMixin,
    ModelFormMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
)


__all__ = [
    "GenericViewSet",
    "ModelViewSet",
    "ReadOnlyModelViewSet",
    "CreateModelMixin",
    "DeleteModelMixin",
    "ListModelMixin",
    "ModelFormMixin",
    "RetrieveModelMixin",
    "UpdateModelMixin",
]



from django.core.exceptions import ValidationError
from django.test import TestCase
from django_sorcery.exceptions import NestedValidationError


class TestNestedValidationError(TestCase):
    def test_string(self):
        error = NestedValidationError("error")
        self.assertEqual(error.messages, ["error"])

        error = NestedValidationError("error %s", params=("here",))
        self.assertEqual(error.messages, ["error here"])

    def test_list(self):
        error = NestedValidationError(["error"])
        self.assertEqual(error.messages, ["error"])

    def test_dict(self):
        error = NestedValidationError({"some": "error"})
        self.assertEqual(error.message_dict, {"some": ["error"]})

    def test_nested(self):
        error = NestedValidationError(NestedValidationError({"some": {"field": ["error"]}}))
        self.assertEqual(error.message_dict, {"some": {"field": ["error"]}})

    def test_update_error_dict(self):
        error = NestedValidationError("error")
        self.assertDictEqual(
            {"some": "other error", "__all__": [error]}, error.update_error_dict({"some": "other error"})
        )

        other = ValidationError("other error")
        error2 = NestedValidationError({"some": error})
        self.assertDictEqual({"some": [other, error]}, error2.update_error_dict({"some": [other]}))

    def test_dict_nested(self):
        error = NestedValidationError({"some": {"field": "error", "list": ["errors"], "object": {"field": "error"}}})
        self.assertEqual(
            error.message_dict, {"some": {"field": ["error"], "list": ["errors"], "object": {"field": ["error"]}}}
        )

    def test_list_nested(self):
        error = NestedValidationError(
            {
                "some": {
                    "field": "error",
                    "list": ["error1", "error2"],
                    "object": {"field": "error"},
                    "nested": [{"field": "error", "list": ["errors"]}],
                }
            }
        )
        self.assertEqual(
            error.message_dict,
            {
                "some": {
                    "field": ["error"],
                    "list": ["error1", "error2"],
                    "object": {"field": ["error"]},
                    "nested": [{"field": ["error"], "list": ["errors"]}],
                }
            },
        )

    def test_nested_error(self):
        error = NestedValidationError({"some": ValidationError("error")})
        self.assertEqual(error.message_dict, {"some": ["error"]})

        inner = NestedValidationError("error")
        error = NestedValidationError({"some": inner})
        self.assertEqual(error.message_dict, {"some": ["error"]})

    def test_repr(self):
        self.assertTrue(repr(NestedValidationError("error")).startswith("NestedValidationError"))



import os

from ... import testapp


MIGRATION_DIR = os.path.abspath(os.path.join(os.path.dirname(testapp.__file__), "migrations"))


class MigrationMixin:
    def write_migration(self, content, filename):
        with open(os.path.join(MIGRATION_DIR, filename), "w+") as migration:
            migration.write(content)

    def delete_migration(self, filename):
        try:
            os.remove(os.path.join(MIGRATION_DIR, filename))
        except OSError:
            pass



"""Revision command."""
from functools import partial

import alembic

from ...db import signals
from ..alembic import AlembicCommand


class Revision(AlembicCommand):
    """Creates an alembic migration revision."""

    help = "Create a migration revision"

    def add_arguments(self, parser):
        parser.add_argument(
            "args", metavar="app_label", nargs=1, help="Specify the app label to create migrations for."
        )
        parser.add_argument(
            "-m", "--message", action="store", required=True, help="Message string to use with 'revision'"
        )
        parser.add_argument(
            "-r", "--rev-id", action="store", help="Specify a hardcoded revision id instead of generating one"
        )
        parser.add_argument(
            "--head", action="store", help="Specify head revision or <branchname>@head to base new revision on."
        )
        parser.add_argument(
            "--branch-label", action="store", help="Specify a branch label to apply to the new revision."
        )
        parser.add_argument(
            "--depends-on",
            action="store",
            help="Specify one or more revision identifiers which this revision should depend on.",
        )
        parser.add_argument(
            "--splice",
            action="store_true",
            default=None,
            help="Allow a non-head revision as the 'head' to splice onto.",
        )
        parser.add_argument(
            "--autogenerate",
            action="store_true",
            dest="autogenerate",
            help="Populate revision script with candidate migration operations, based on comparison of database to model.",
        )
        parser.add_argument(
            "--no-autogenerate",
            action="store_false",
            dest="autogenerate",
            default=True,
            help="Generate blank candidate migration. This will not compare database to models.",
        )

    def handle(
        self,
        app_label,
        message=None,
        head=None,
        splice=None,
        branch_label=None,
        depends_on=None,
        rev_id=None,
        autogenerate=None,
        **kwargs,
    ):
        appconfig = self.lookup_app(app_label)

        version_tables = {
            ".".join(
                filter(
                    None,
                    [
                        appconf.config.get_main_option("version_table_schema"),
                        appconf.config.get_main_option("version_table"),
                    ],
                )
            )
            for appconf in self.sorcery_apps.values()
        }

        @signals.alembic_include_object.connect
        def include_object(obj=None, name=None, type_=None, reflected=None, compare_to=None):
            if type_ == "table":
                return obj in appconfig.tables and obj.fullname not in version_tables

            else:
                return obj.table in appconfig.tables

        command_args = {
            "autogenerate": autogenerate,
            "branch_label": branch_label,
            "depends_on": depends_on,
            "head": head,
            "rev_id": rev_id,
            "message": message,
            "splice": splice,
            "sql": False,
            "version_path": appconfig.version_path,
        }
        self.revision_context = alembic.autogenerate.RevisionContext(appconfig.config, appconfig.script, command_args)
        with alembic.context.EnvironmentContext(
            appconfig.config,
            appconfig.script,
            fn=partial(self.generate_migration, appconfig=appconfig),
            as_sql=False,
            template_args=self.revision_context.template_args,
            revision_context=self.revision_context,
        ) as context:
            self.run_env(context, appconfig)

        for _ in self.revision_context.generate_scripts():
            pass  # iterate over to generate the migration scripts

    def generate_migration(self, rev, context, appconfig=None):
        """Generate alembic migration."""
        self.revision_context.run_autogenerate(rev, context)
        return []


Command = Revision



import json

from django.core.exceptions import ValidationError
from django_sorcery import fields
from django_sorcery.forms import (
    apply_limit_choices_to_form_field,
    modelform_factory,
)

from .base import TestCase
from .testapp.models import CompositePkModel, Owner, Vehicle, VehicleType, db


class TestEnumField(TestCase):
    def test_field(self):
        field = fields.EnumField(enum_class=VehicleType)

        with self.assertRaises(ValidationError) as ctx:
            value = field.clean(None)
        self.assertEqual(ctx.exception.code, "required")

        with self.assertRaises(ValidationError) as ctx:
            value = field.clean("")
        self.assertEqual(ctx.exception.code, "required")
        value = field.to_python("")

        value = field.clean("car")
        self.assertEqual(value, VehicleType.car)

        value = field.clean("Car")
        self.assertEqual(value, VehicleType.car)

        value = field.to_python(VehicleType.car)
        self.assertEqual(value, VehicleType.car)

        with self.assertRaises(ValidationError):
            field.clean("blue")

        value = field.valid_value(None)
        self.assertFalse(value)

        value = field.prepare_value(VehicleType.car)
        self.assertEqual(value, "car")

        value = field.bound_data(VehicleType.car, None)
        self.assertEqual(value, "car")

        value = field.bound_data(None, None)
        self.assertIsNone(value)

        value = field.bound_data(None, VehicleType.car)
        self.assertIsNone(value, "car")

        value = field.prepare_value(None)
        self.assertFalse(value)

    def test_field_not_required(self):
        field = fields.EnumField(enum_class=VehicleType, required=False)

        value = field.clean(None)
        self.assertIsNone(value)

        value = field.clean("")
        self.assertIsNone(value)

        value = field.clean("car")
        self.assertEqual(value, VehicleType.car)

        value = field.clean("Car")
        self.assertEqual(value, VehicleType.car)

        with self.assertRaises(ValidationError):
            field.to_python("blue")

        value = field.bound_data(VehicleType.car, None)
        self.assertEqual(value, "car")

        value = field.bound_data(None, None)
        self.assertIsNone(value)

        value = field.bound_data(None, VehicleType.car)
        self.assertIsNone(value, "car")


class TestModelChoiceField(TestCase):
    def setUp(self):
        super().setUp()
        db.add_all([Owner(first_name=f"first_name {i}", last_name=f"last_name {i}") for i in range(10)])
        db.flush()

    def test_apply_limit_value(self):
        field = fields.ModelChoiceField(Owner, db, limit_choices_to=[Owner.first_name == "first_name 1"])
        apply_limit_choices_to_form_field(field)
        self.assertEqual(field.queryset.count(), 1)

    def test_apply_limit_callable(self):
        def limit_choices_to():
            return [Owner.first_name == "first_name 1"]

        field = fields.ModelChoiceField(Owner, db, required=True, limit_choices_to=limit_choices_to)
        apply_limit_choices_to_form_field(field)
        self.assertEqual(field.queryset.count(), 1)

    def test_choices(self):
        field = fields.ModelChoiceField(Owner, db)

        self.assertListEqual(
            list(field.choices), [("", field.empty_label)] + [(owner.id, str(owner)) for owner in Owner.query]
        )

        field = fields.ModelChoiceField(Owner, db, required=True, initial=1)
        self.assertListEqual(list(field.choices), [(owner.id, str(owner)) for owner in Owner.query])

    def test_get_object(self):
        owner = Owner.objects.first()

        field = fields.ModelChoiceField(Owner, db)

        self.assertIsNone(field.get_object(None))

        owner = field.get_object(owner.id)
        self.assertIsNotNone(owner)
        self.assertIsInstance(owner, Owner)

        with self.assertRaises(ValidationError) as ctx:
            field.get_object(0)

        self.assertEqual(
            ctx.exception.args,
            ("Select a valid choice. That choice is not one of the available choices.", "invalid_choice", None),
        )

    def test_get_object_composite_pk(self):
        instance = CompositePkModel(id=1, pk=2)
        db.add(instance)
        db.flush()

        field = fields.ModelChoiceField(CompositePkModel, db)
        pks = field.prepare_value(instance)

        obj = field.get_object(json.dumps(pks))
        self.assertIsNotNone(obj)
        self.assertIs(obj, instance)

    def test_to_python(self):
        owner = Owner.objects.first()
        field = fields.ModelChoiceField(Owner, db)
        owner = field.to_python(owner.id)
        self.assertIsNotNone(owner)
        self.assertIsInstance(owner, Owner)

    def test_label_from_instance(self):
        owner = Owner.objects.first()

        field = fields.ModelChoiceField(Owner, db)

        self.assertEqual(field.label_from_instance(Owner.query.get(owner.id)), repr(owner))

    def test_prepare_instance_value(self):
        field = fields.ModelChoiceField(Owner, db)
        owner = Owner.objects.first()

        pks = field.prepare_instance_value(Owner.query.get(owner.id))
        self.assertEqual(pks, owner.id)

    def test_prepare_instance_value_composite(self):
        field = fields.ModelChoiceField(CompositePkModel, db)

        pks = field.prepare_instance_value(CompositePkModel(id=1, pk="a"))
        self.assertDictEqual(pks, {"id": 1, "pk": "a"})

    def test_prepare_value(self):
        field = fields.ModelChoiceField(CompositePkModel, db)
        pks = field.prepare_value(CompositePkModel(id=1, pk="a"))
        self.assertDictEqual(pks, {"id": 1, "pk": "a"})

    def test_validate(self):
        field = fields.ModelChoiceField(Owner, db, required=True)

        self.assertIsNone(field.validate(1))

        with self.assertRaises(ValidationError):
            field.validate(None)

    def test_get_bound_field(self):
        db.rollback()
        form = modelform_factory(Vehicle, fields=("owner",), session=db)()
        field = form.fields["owner"]
        bf = field.get_bound_field(form, "owner")
        self.assertHTMLEqual(
            str(bf), '<select name="owner" id="id_owner"><option value="" selected>---------</option></select>'
        )


class TestModelMultipleChoiceField(TestCase):
    def setUp(self):
        super().setUp()
        db.add_all([Owner(first_name=f"first_name {i}", last_name=f"last_name {i}") for i in range(10)])
        db.flush()

    def test_to_python(self):
        field = fields.ModelMultipleChoiceField(Owner, db)
        owner1, owner2, owner3 = Owner.objects[:3]

        self.assertEqual(field.to_python(None), [])

        self.assertEqual(field.to_python([owner1.id, owner2.id, owner3.id]), [owner1, owner2, owner3])

    def test_prepare_value(self):
        field = fields.ModelMultipleChoiceField(Owner, db)
        owner1, owner2, owner3 = Owner.objects[:3]

        self.assertIsNone(field.prepare_value(None))

        self.assertEqual(field.prepare_value([owner1, owner2, owner3]), [owner1.id, owner2.id, owner3.id])



import os

import six

from django.test import TestCase
from django_sorcery.db import databases
from django_sorcery.management.commands.sorcery_downgrade import Downgrade
from django_sorcery.management.commands.sorcery_revision import Command
from django_sorcery.management.commands.sorcery_upgrade import Upgrade

from .base import MIGRATION_DIR, MigrationMixin


rev = "000000000000"

db = databases.get("test")


class TestRevision(MigrationMixin, TestCase):
    def setUp(self):
        super().setUp()
        Upgrade().run_from_argv(["./manage.py sorcery", "upgrade", "--no-color"])

    def tearDown(self):
        super().tearDown()
        self.delete_migration("{}_.py".format(rev))
        self.delete_migration("{}_zero.py".format(rev))
        Downgrade().run_from_argv(["./manage.py sorcery", "downgrade", "--no-color"])

    def test_without_app(self):
        out = six.StringIO()

        cmd = Command(stdout=out)

        with self.assertRaises(SystemExit):
            cmd.run_from_argv(["./manage.py sorcery", "makemigrations", "--no-color"])

    def test_with_bad_app(self):
        out = six.StringIO()

        cmd = Command(stdout=out)

        with self.assertRaises(SystemExit):
            cmd.run_from_argv(["./manage.py sorcery", "makemigrations", "dumdum", "--no-color"])

    def test_with_name(self):
        out = six.StringIO()

        cmd = Command(stdout=out)

        cmd.run_from_argv(
            ["./manage.py sorcery", "makemigrations", "tests_testapp", "-r", rev, "-m", "zero", "--no-color"]
        )

        self.assertTrue(os.path.isfile(os.path.join(MIGRATION_DIR, "{}_zero.py".format(rev))))

    def test_longer_version_table_identifier(self):
        out = six.StringIO()
        err = six.StringIO()

        cmd = Command(stdout=out, stderr=err)

        original_length = db.engine.dialect.max_identifier_length
        db.engine.dialect.max_identifier_length = 5

        with self.assertRaises(SystemExit):
            cmd.run_from_argv(
                ["./manage.py sorcery", "makemigrations", "tests_testapp", "-r", rev, "-m", "zero", "--no-color"]
            )

        db.engine.dialect.max_identifier_length = original_length

        err.seek(0)
        self.assertEqual(
            err.readlines(),
            [
                "CommandError: 'alembic_version_tests_testapp' is 29 characters long which "
                "is an invalid identifier in 'postgresql' as its max idenfier length is 5\n"
            ],
        )



from django_sorcery.db.composites import CompositeField

from ..base import TestCase
from ..testapp.models import Address, Business, States, db


class TestComposite(TestCase):
    def test_autogenerate_columns(self):
        self.assertIsInstance(Business.location.property, CompositeField)
        self.assertIsInstance(Business.other_location.property, CompositeField)

        location_props = [p.key for p in Business.location.property.props]
        other_location_props = [p.key for p in Business.other_location.property.props]

        self.assertListEqual(["_location_state", "_location_street", "_location_zip"], location_props)
        self.assertListEqual(["_foo_state", "_foo_street", "_foo_zip"], other_location_props)

        self.assertEqual(Business._location_state.prop.columns[0].type.name, "states")
        self.assertEqual(Business._foo_state.prop.columns[0].type.name, "foo_states")

    def test_can_persist(self):
        instance = Business()
        instance.location = Address(States.NY, "street", "123")
        instance.other_location = Address(street="other street", state=States.NJ, zip="456")
        db.add(instance)
        db.flush()
        db.expire_all()

        instance = Business.objects.first()

        self.assertEqual(set(instance.location.__composite_values__()), {"street", States.NY, "123"})
        self.assertEqual(set(instance.other_location.__composite_values__()), {"other street", States.NJ, "456"})

    def test_repr(self):
        self.assertEqual(
            repr(Address(street="other street", state="NJ", zip="456")),
            "Address(state='NJ', street='other street', zip='456')",
        )
        self.assertEqual(repr(Address(state="NJ", zip="456")), "Address(state='NJ', zip='456')")

    def test_eq(self):
        self.assertEqual(
            Address(street="other street", state="NJ", zip="456"),
            Address(street="other street", state="NJ", zip="456"),
        )
        self.assertNotEqual(
            Address(street="other street", state="NJ", zip="456"),
            Address(street="another street", state="NJ", zip="456"),
        )
        self.assertFalse(Address(street="other street", state="NJ", zip="456").__eq__(None))

    def test_bool(self):
        self.assertTrue(Address(street="street"))
        self.assertFalse(Address())



import six

from django.test import TestCase
from django_sorcery.db import databases
from django_sorcery.management.commands.sorcery_downgrade import Downgrade
from django_sorcery.management.commands.sorcery_upgrade import Command

from .base import MigrationMixin


MIGRATION = '''"""zero

Revision ID: {rev}
Revises:
Create Date: 2018-07-24 01:28:40.893136

"""


# revision identifiers, used by Alembic.
revision = '{rev}'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
'''

db = databases.get("test")


class TestUpgrade(MigrationMixin, TestCase):
    def setUp(self):
        super().setUp()
        for rev in ["000000000000", "000000000001"]:
            self.write_migration(MIGRATION.format(rev=rev), "{}_zero.py".format(rev))
            self.write_migration(MIGRATION.format(rev=rev), "{}_zero.py".format(rev))
        Downgrade().run_from_argv(["./manage.py sorcery", "downgrade", "tests_testapp", "--no-color"])

    def tearDown(self):
        super().tearDown()
        Downgrade().run_from_argv(["./manage.py sorcery", "downgrade", "tests_testapp", "--no-color"])
        for rev in ["000000000000", "000000000001"]:
            self.delete_migration("{}_zero.py".format(rev))
        Downgrade().run_from_argv(["./manage.py sorcery", "downgrade", "tests_testapp", "--no-color"])

    def test(self):
        out = six.StringIO()
        cmd = Command(stdout=out)
        cmd.run_from_argv(["./manage.py sorcery", "upgrade", "--no-color"])

        revs = db.execute("select * from alembic_version_tests_testapp").fetchall()
        self.assertEqual({r.version_num for r in revs}, {"000000000001", "000000000000"})

    def test_sql(self):
        out = six.StringIO()
        cmd = Command(stdout=out)
        cmd.run_from_argv(["./manage.py sorcery", "upgrade", "-s", "--no-color"])

        out.seek(0)
        lines = out.readlines()

        self.assertEqual(len(lines), 23)
        statements = [
            "-- Running upgrade  -> 000000000001\n",
            "INSERT INTO public.alembic_version_tests_testapp (version_num) VALUES ('000000000001')",
            "-- Running upgrade  -> 000000000000\n",
            "INSERT INTO public.alembic_version_tests_testapp (version_num) VALUES ('000000000000')",
        ]
        for statement in statements:
            found = any(statement in line for line in lines)

            self.assertTrue(found, statement)

    def test_with_range_no_app(self):
        out = six.StringIO()
        cmd = Command(stdout=out)

        with self.assertRaises(SystemExit):
            cmd.run_from_argv(["./manage.py sorcery", "upgrade", "-r", "000000000001", "--no-color"])

    def test_with_range(self):
        out = six.StringIO()
        cmd = Command(stdout=out)

        cmd.run_from_argv(["./manage.py sorcery", "upgrade", "tests_testapp", "-r", ":000000000000", "--no-color"])
        revs = db.execute("select * from public.alembic_version_tests_testapp").fetchall()
        self.assertEqual(revs, [("000000000000",)])

    def test_catching_alembic_error(self):
        out = six.StringIO()
        err = six.StringIO()

        cmd = Command(stdout=out, stderr=err)

        with self.assertRaises(SystemExit):
            cmd.run_from_argv(["./manage.py sorcery", "upgrade", "tests_testapp", "-r", "revision", "--no-color"])

        err.seek(0)
        self.assertEqual(err.readlines(), ["CommandError: Can't locate revision identified by 'revision'\n"])



from django.core.exceptions import ImproperlyConfigured
from django_sorcery.forms import ALL_FIELDS
from django_sorcery.formsets import modelformset_factory

from ..base import TestCase
from ..testapp.models import Owner, db


class TestModelFormSet(TestCase):
    def setUp(self):
        super().setUp()
        self.owners = [
            Owner(first_name="Test 1", last_name="Owner 1"),
            Owner(first_name="Test 2", last_name="Owner 2"),
            Owner(first_name="Test 3", last_name="Owner 3"),
            Owner(first_name="Test 4", last_name="Owner 4"),
        ]
        db.add_all(self.owners)
        db.flush()

    def test_factory(self):
        formset_class = modelformset_factory(Owner, fields=ALL_FIELDS, session=db)
        formset = formset_class()

        self.assertEqual(formset.model, Owner)
        self.assertEqual(formset.session, db)

    def test_render(self):
        formset_class = modelformset_factory(Owner, fields=("first_name", "last_name"), session=db)
        query = Owner.query

        formset = formset_class(queryset=query)

        self.assertEqual(len(formset.forms), 5)

        for form in formset.forms:
            form.order_fields(sorted(form.fields.keys()))

        soup = self.get_soup(formset.as_p())
        expected_soup = self.get_soup(
            "".join(
                [
                    '<input type="hidden" name="form-TOTAL_FORMS" value="5" id="id_form-TOTAL_FORMS" />',
                    '<input type="hidden" name="form-INITIAL_FORMS" value="4" id="id_form-INITIAL_FORMS" />',
                    '<input type="hidden" name="form-MIN_NUM_FORMS" value="0" id="id_form-MIN_NUM_FORMS" />',
                    '<input type="hidden" name="form-MAX_NUM_FORMS" value="1000" id="id_form-MAX_NUM_FORMS" />',
                    "<p>",
                    '  <label for="id_form-0-first_name">First name:</label>',
                    '  <input type="text" name="form-0-first_name" id="id_form-0-first_name" maxlength="50" value="Test 1" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-0-last_name">Last name:</label>',
                    '  <input type="text" name="form-0-last_name" id="id_form-0-last_name" maxlength="50" value="Owner 1" />',
                    '  <input type="text" name="form-0-id" id="id_form-0-id" type="hidden" value="{}" />'.format(
                        self.owners[0].id
                    ),
                    "</p>",
                    "<p>",
                    '  <label for="id_form-1-first_name">First name:</label>',
                    '  <input type="text" name="form-1-first_name" id="id_form-1-first_name" maxlength="50" value="Test 2" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-1-last_name">Last name:</label>',
                    '  <input type="text" name="form-1-last_name" id="id_form-1-last_name" maxlength="50" value="Owner 2" />',
                    '  <input type="text" name="form-1-id" id="id_form-1-id" type="hidden" value="{}" />'.format(
                        self.owners[1].id
                    ),
                    "</p>",
                    "<p>",
                    '  <label for="id_form-2-first_name">First name:</label>',
                    '  <input type="text" name="form-2-first_name" id="id_form-2-first_name" maxlength="50" value="Test 3" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-2-last_name">Last name:</label>',
                    '  <input type="text" name="form-2-last_name" id="id_form-2-last_name" maxlength="50" value="Owner 3" />',
                    '  <input type="text" name="form-2-id" id="id_form-2-id" type="hidden" value="{}" />'.format(
                        self.owners[2].id
                    ),
                    "</p>",
                    "<p>",
                    '  <label for="id_form-3-first_name">First name:</label>',
                    '  <input type="text" name="form-3-first_name" id="id_form-3-first_name" maxlength="50" value="Test 4" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-3-last_name">Last name:</label>',
                    '  <input type="text" name="form-3-last_name" id="id_form-3-last_name" maxlength="50" value="Owner 4" />',
                    '  <input type="text" name="form-3-id" id="id_form-3-id" type="hidden" value="{}" />'.format(
                        self.owners[3].id
                    ),
                    "</p>",
                    "<p>",
                    '  <label for="id_form-4-first_name">First name:</label>',
                    '  <input type="text" name="form-4-first_name" id="id_form-4-first_name" maxlength="50" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-4-last_name">Last name:</label>',
                    '  <input type="text" name="form-4-last_name" id="id_form-4-last_name" maxlength="50" />',
                    "</p>",
                ]
            )
        )
        self.assertHTMLEqual(soup.prettify(), expected_soup.prettify())

    def test_edit(self):
        formset_class = modelformset_factory(Owner, fields=("first_name", "last_name"), session=db)
        query = Owner.query

        data = {
            "form-TOTAL_FORMS": "5",
            "form-INITIAL_FORMS": "4",
            "form-MIN_NUM_FORMS": "0",
            "form-MAX_NUM_FORMS": "1000",
            "form-0-id": self.owners[0].id,
            "form-0-first_name": "Edited first name {}".format(self.owners[0].id),
            "form-0-last_name": "Edited last name {}".format(self.owners[0].id),
            "form-1-id": self.owners[1].id,
            "form-1-first_name": "Edited first name {}".format(self.owners[1].id),
            "form-1-last_name": "Edited last name {}".format(self.owners[1].id),
            "form-2-id": self.owners[2].id,
            "form-2-first_name": "Edited first name {}".format(self.owners[2].id),
            "form-2-last_name": "Edited last name {}".format(self.owners[2].id),
            "form-3-id": self.owners[3].id,
            "form-3-first_name": "Edited first name {}".format(self.owners[3].id),
            "form-3-last_name": "Edited last name {}".format(self.owners[3].id),
            "form-4-id": "",
            "form-4-first_name": "New first name",
            "form-4-last_name": "New last name",
        }

        formset = formset_class(queryset=query, data=data)
        self.assertTrue(formset.is_valid())
        instances = formset.save()
        db.expire_all()

        for owner in instances:
            if owner.first_name.startswith("Edited"):
                self.assertEqual(owner.first_name, "Edited first name {}".format(owner.id))
                self.assertEqual(owner.last_name, "Edited last name {}".format(owner.id))
            else:
                self.assertEqual(owner.first_name, "New first name")
                self.assertEqual(owner.last_name, "New last name")

    def test_edit_new_delete_ignore(self):
        formset_class = modelformset_factory(Owner, fields=("first_name", "last_name"), session=db, can_delete=True)

        data = {
            "form-TOTAL_FORMS": "1",
            "form-INITIAL_FORMS": "0",
            "form-MIN_NUM_FORMS": "0",
            "form-MAX_NUM_FORMS": "1000",
            "form-0-first_name": "Edited first name 1",
            "form-0-last_name": "Edited last name 1",
            "form-0-DELETE": "on",
        }

        formset = formset_class(queryset=[], data=data)
        self.assertTrue(formset.is_valid())
        instances = formset.save()
        self.assertEqual(instances, [])
        self.assertEqual(formset.new_objects, [])
        self.assertEqual(formset.changed_objects, [])
        self.assertEqual(formset.deleted_objects, [])

    def test_delete_render(self):
        formset_class = modelformset_factory(Owner, fields=("first_name", "last_name"), session=db, can_delete=True)
        formset = formset_class(queryset=Owner.query)

        self.assertEqual(len(formset.forms), 5)

        for form in formset.forms:
            form.order_fields(sorted(form.fields.keys()))

        soup = self.get_soup(formset.as_p())
        expected_soup = self.get_soup(
            "".join(
                [
                    '<input type="hidden" name="form-TOTAL_FORMS" value="5" id="id_form-TOTAL_FORMS" />',
                    '<input type="hidden" name="form-INITIAL_FORMS" value="4" id="id_form-INITIAL_FORMS" />',
                    '<input type="hidden" name="form-MIN_NUM_FORMS" value="0" id="id_form-MIN_NUM_FORMS" />',
                    '<input type="hidden" name="form-MAX_NUM_FORMS" value="1000" id="id_form-MAX_NUM_FORMS" />',
                    "<p>",
                    '  <label for="id_form-0-DELETE">Delete:</label>',
                    '  <input id="id_form-0-DELETE" name="form-0-DELETE" type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-0-first_name">First name:</label>',
                    '  <input type="text" name="form-0-first_name" id="id_form-0-first_name" maxlength="50" value="Test 1" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-0-last_name">Last name:</label>',
                    '  <input type="text" name="form-0-last_name" id="id_form-0-last_name" maxlength="50" value="Owner 1" />',
                    '  <input type="text" name="form-0-id" id="id_form-0-id" type="hidden" value="{}" />'.format(
                        self.owners[0].id
                    ),
                    "</p>",
                    "<p>",
                    '  <label for="id_form-1-DELETE">Delete:</label>',
                    '  <input id="id_form-1-DELETE" name="form-1-DELETE" type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-1-first_name">First name:</label>',
                    '  <input type="text" name="form-1-first_name" id="id_form-1-first_name" maxlength="50" value="Test 2" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-1-last_name">Last name:</label>',
                    '  <input type="text" name="form-1-last_name" id="id_form-1-last_name" maxlength="50" value="Owner 2" />',
                    '  <input type="text" name="form-1-id" id="id_form-1-id" type="hidden" value="{}" />'.format(
                        self.owners[1].id
                    ),
                    "</p>",
                    "<p>",
                    '  <label for="id_form-2-DELETE">Delete:</label>',
                    '  <input id="id_form-2-DELETE" name="form-2-DELETE" type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-2-first_name">First name:</label>',
                    '  <input type="text" name="form-2-first_name" id="id_form-2-first_name" maxlength="50" value="Test 3" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-2-last_name">Last name:</label>',
                    '  <input type="text" name="form-2-last_name" id="id_form-2-last_name" maxlength="50" value="Owner 3" />',
                    '  <input type="text" name="form-2-id" id="id_form-2-id" type="hidden" value="{}" />'.format(
                        self.owners[2].id
                    ),
                    "</p>",
                    "<p>",
                    '  <label for="id_form-3-DELETE">Delete:</label>',
                    '  <input id="id_form-3-DELETE" name="form-3-DELETE" type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-3-first_name">First name:</label>',
                    '  <input type="text" name="form-3-first_name" id="id_form-3-first_name" maxlength="50" value="Test 4" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-3-last_name">Last name:</label>',
                    '  <input type="text" name="form-3-last_name" id="id_form-3-last_name" maxlength="50" value="Owner 4" />',
                    '  <input type="text" name="form-3-id" id="id_form-3-id" type="hidden" value="{}" />'.format(
                        self.owners[3].id
                    ),
                    "</p>",
                    "<p>",
                    '  <label for="id_form-4-DELETE">Delete:</label>',
                    '  <input id="id_form-4-DELETE" name="form-4-DELETE" type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-4-first_name">First name:</label>',
                    '  <input type="text" name="form-4-first_name" id="id_form-4-first_name" maxlength="50" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_form-4-last_name">Last name:</label>',
                    '  <input type="text" name="form-4-last_name" id="id_form-4-last_name" maxlength="50" />',
                    "</p>",
                ]
            )
        )
        self.assertHTMLEqual(soup.prettify(), expected_soup.prettify())

    def test_delete(self):
        formset_class = modelformset_factory(Owner, fields=("first_name", "last_name"), session=db, can_delete=True)
        query = Owner.query

        data = {
            "form-TOTAL_FORMS": "5",
            "form-INITIAL_FORMS": "4",
            "form-MIN_NUM_FORMS": "0",
            "form-MAX_NUM_FORMS": "1000",
            "form-0-id": self.owners[0].id,
            "form-0-first_name": "Edited first name {}".format(self.owners[0].id),
            "form-0-last_name": "Edited last name {}".format(self.owners[0].id),
            "form-1-id": self.owners[1].id,
            "form-1-first_name": "Edited first name {}".format(self.owners[1].id),
            "form-1-last_name": "Edited last name {}".format(self.owners[1].id),
            "form-2-id": self.owners[2].id,
            "form-2-first_name": "Edited first name {}".format(self.owners[2].id),
            "form-2-last_name": "Edited last name {}".format(self.owners[2].id),
            "form-3-id": self.owners[3].id,
            "form-3-first_name": "Edited first name {}".format(self.owners[3].id),
            "form-3-last_name": "Edited last name {}".format(self.owners[3].id),
            "form-3-DELETE": "on",
            "form-4-id": "",
            "form-4-first_name": "",
            "form-4-last_name": "",
        }

        formset = formset_class(queryset=query, data=data)
        self.assertTrue(formset.is_valid())
        instances = formset.save()

        self.assertEqual(len(instances), 3)
        self.assertEqual(len(formset.deleted_objects), 1)
        self.assertEqual(Owner.query.count(), 3)

    def test_initial_extra(self):
        formset_class = modelformset_factory(Owner, fields="__all__", session=db)

        initial = [{"last_name": "test"}, None]

        formset = formset_class(queryset=Owner.query, initial=initial)
        extra_form = formset.forms[formset.initial_form_count()]

        self.assertEqual(extra_form.initial["last_name"], "test")

    def test_bad_initial(self):
        formset_class = modelformset_factory(Owner, fields="__all__", session=db, extra=2)

        initial = [None]

        formset = formset_class(queryset=Owner.query, initial=initial)
        extra_form = formset.forms[formset.initial_form_count() + 1]

        self.assertEqual(extra_form.initial["first_name"], None)
        self.assertEqual(extra_form.initial["last_name"], None)

    def test_no_queryset(self):
        formset_class = modelformset_factory(Owner, fields=("first_name", "last_name"), session=db)
        formset = formset_class()

        self.assertIsNone(formset.queryset)
        self.assertListEqual(formset.get_queryset(), Owner.query.all())

    def test_save_no_initial_forms(self):
        formset_class = modelformset_factory(Owner, fields=("first_name", "last_name"), session=db, extra=0)
        formset = formset_class(queryset=Owner.query.filter_by(id=99))
        self.assertEqual(formset.save(), [])

    def test_formset_factory_no_fields_exclude(self):
        with self.assertRaises(ImproperlyConfigured) as ctx:
            modelformset_factory(Owner, fields=None, exclude=None, session=db)

        self.assertEqual(
            ctx.exception.args,
            ("Calling modelformset_factory without defining 'fields' or " "'exclude' explicitly is prohibited.",),
        )



import sqlalchemy as sa
from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from django.utils.html import escape
from django_sorcery.views.list import (
    MultipleObjectMixin,
    MultipleObjectTemplateResponseMixin,
)

from ..base import TestCase
from ..testapp.models import Owner, db


class TestListView(TestCase):
    def setUp(self):
        super().setUp()
        db.add_all(
            [
                Owner(id=1, first_name="Test 1", last_name="Owner 1"),
                Owner(id=2, first_name="Test 2", last_name="Owner 2"),
                Owner(id=3, first_name="Test 3", last_name="Owner 3"),
                Owner(id=4, first_name="Test 4", last_name="Owner 4"),
            ]
        )
        db.flush()

    def test_paginate_orphans(self):
        view = MultipleObjectMixin()
        self.assertEqual(view.get_paginate_orphans(), view.paginate_orphans)

    def test_get_template_names(self):
        view = MultipleObjectTemplateResponseMixin()

        with self.assertRaises(ImproperlyConfigured):
            view.get_template_names()

    def test_list_generic(self):
        url = reverse("owners_list")

        response = self.client.get(url)

        self.assertEqual(
            response.content.decode().strip(), "owner" + escape("".join([repr(owner) for owner in Owner.query]))
        )

    def test_list_with_template(self):
        url = reverse("owners_list_tmpl")

        response = self.client.get(url)

        self.assertEqual(response.content.decode().strip(), escape("".join([repr(owner) for owner in Owner.query])))

    def test_list_ordering(self):
        url = reverse("owners_list_order")

        response = self.client.get(url)

        owners = Owner.query.order_by(Owner.id.desc())

        self.assertEqual(
            response.content.decode().strip(), "owner" + escape("".join([repr(owner) for owner in owners]))
        )

    def test_list_pagination(self):
        db.add_all([Owner(id=i, first_name=f"Test {i}", last_name=f"Owner {i}") for i in range(5, 50)])
        db.flush()

        url = reverse("owners_list")

        response = self.client.get(url)

        self.assertEqual(
            response.content.decode().strip(),
            "owner" + escape("".join([repr(owner) for owner in Owner.query.order_by(Owner.id)[:20]])),
        )

    def test_list_pagination_last_page(self):
        db.add_all([Owner(id=i, first_name=f"Test {i}", last_name=f"Owner {i}") for i in range(5, 50)])
        db.flush()

        url = reverse("owners_list")

        response = self.client.get(f"{url}?page=last")

        self.assertEqual(
            response.content.decode().strip(),
            "owner" + escape("".join([repr(owner) for owner in Owner.query.order_by(Owner.id)[40:]])),
        )

    def test_list_pagination_bad_page(self):
        db.add_all([Owner(id=i, first_name=f"Test {i}", last_name=f"Owner {i}") for i in range(5, 50)])
        db.flush()

        url = reverse("owners_list")

        response = self.client.get(f"{url}?page=abc")

        self.assertEqual(response.status_code, 404)

    def test_list_pagination_bad_page_out_of_range(self):
        db.add_all([Owner(id=i, first_name=f"Test {i}", last_name=f"Owner {i}") for i in range(5, 50)])
        db.flush()

        url = reverse("owners_list")

        response = self.client.get(f"{url}?page=100")

        self.assertEqual(response.status_code, 404)

    def test_list_context_name(self):
        url = reverse("owners_list_context_name")
        response = self.client.get(url)
        self.assertIn("owner_list", response.context_data)
        self.assertIsInstance(response.context_data["owner_list"], sa.orm.Query)

    def test_list_no_empty_paginate(self):
        db.rollback()

        url = reverse("owners_list_no_empty_paginate")
        response = self.client.get(url)

        self.assertEqual(response.status_code, 404)

    def test_list_dummy_no_empty_paginate(self):
        db.rollback()

        url = reverse("dummy_list_no_empty_paginate")
        response = self.client.get(url)

        self.assertEqual(response.status_code, 404)



from django.test import TestCase
from django_sorcery.management.commands.sorcery import Command

from ...models_multidb import default_db, other_db


class TestCreateAllDropAll(TestCase):
    def setUp(self):
        super().setUp()
        default_db.drop_all()
        other_db.drop_all()

    def tearDown(self):
        super().setUp()
        default_db.create_all()
        other_db.create_all()

    def test_create_drop_commands(self):
        cmd = Command()

        cmd.run_from_argv(["./manage.py", "sorcery", "createall", "-d", other_db.alias])
        cmd.run_from_argv(["./manage.py", "sorcery", "dropall", "-d", default_db.alias])

        self.assertEqual(default_db.inspector.get_table_names(), [])
        self.assertEqual(other_db.inspector.get_table_names(), ["bar"])

        cmd.run_from_argv(["./manage.py", "sorcery", "dropall", "-d", other_db.alias])
        cmd.run_from_argv(["./manage.py", "sorcery", "createall", "-d", default_db.alias])

        self.assertEqual(default_db.inspector.get_table_names(), ["foo"])
        self.assertEqual(other_db.inspector.get_table_names(), [])

        cmd.run_from_argv(["./manage.py", "sorcery", "createall", "-d", other_db.alias])
        cmd.run_from_argv(["./manage.py", "sorcery", "dropall", "-d", default_db.alias])

        self.assertEqual(default_db.inspector.get_table_names(), [])
        self.assertEqual(other_db.inspector.get_table_names(), ["bar"])



"""Base model view things with sqlalchemy."""
from contextlib import suppress

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import InvalidPage, Paginator
from django.http import Http404
from django.utils.translation import gettext
from django.views.generic.base import ContextMixin
from sqlalchemy import literal
from sqlalchemy.exc import InvalidRequestError

from ..db import meta


class SQLAlchemyMixin(ContextMixin):
    """Provides sqlalchemy model view support like query, model, session,
    etc.."""

    queryset = None
    model = None
    session = None
    context_object_name = None
    query_options = None

    @classmethod
    def get_model(cls):
        """Returns the model class."""
        if cls.model is not None:
            return cls.model

        with suppress(AttributeError, InvalidRequestError):
            return cls.queryset._only_full_mapper_zero("get").class_

        raise ImproperlyConfigured(
            "Couldn't figure out the model for %(cls)s, either provide a queryset or set the model "
            "and the session attribute" % {"cls": cls.__name__}
        )

    def get_queryset(self):
        """Return the `QuerySet` that will be used to look up the object.

        This method is called by the default implementation of
        get_object() and may not be called if get_object() is
        overridden.
        """
        if self.queryset is not None:
            return self.queryset

        model = self.get_model()
        query = None
        if model:
            query = getattr(model, "query", None) or getattr(model, "objects", None)

            if query is None and self.session:
                query = self.session.query(model)

        if query:
            query = query.options(*self.get_query_options())

        if not query:
            raise ImproperlyConfigured(
                "%(cls)s is missing a QuerySet. Define %(cls)s.model and %(cls)s.session, %(cls)s.queryset, "
                "or override %(cls)s.get_queryset()." % {"cls": self.__class__.__name__}
            )

        return query

    def get_session(self):
        """Returns the sqlalchemy session."""
        if self.session is None:
            self.session = self.get_queryset().session

        return self.session

    def get_query_options(self):
        """Returns sqlalchemy query options."""
        return self.query_options or []

    def get_model_template_name(self):
        """Returns the base template path."""
        model = self.get_model()
        app_config = apps.get_containing_app_config(model.__module__)
        prefix = model.__name__.lower() if app_config is None else app_config.label
        return f"{prefix}/{model.__name__.lower()}{self.template_name_suffix}.html"


class BaseMultipleObjectMixin(SQLAlchemyMixin):
    """Provides sqlalchemy support for list views."""

    allow_empty = True
    paginate_by = None
    paginate_orphans = 0
    paginator_class = Paginator
    page_kwarg = "page"
    ordering = None

    def get_queryset(self):
        """Return the list of items for this view.

        The return value must be an iterable and may be an instance of
        `QuerySet` in which case `QuerySet` specific behavior will be
        enabled.
        """
        queryset = super().get_queryset()

        ordering = self.get_ordering()
        if ordering is not None:
            queryset = queryset.order_by(*ordering)

        allow_empty = self.get_allow_empty()

        if not allow_empty:
            # When pagination is enabled and object_list is a queryset,
            # it's better to do a cheap query than to load the unpaginated
            # queryset in memory.
            is_empty = True
            if self.get_paginate_by(queryset) is not None:
                if hasattr(queryset, "session"):
                    session = queryset.session
                    is_empty = not session.query(literal(True)).filter(queryset.exists()).scalar()

                else:
                    is_empty = not next(iter(queryset), None)

            if is_empty:
                raise Http404(
                    gettext("Empty list and '%(class_name)s.allow_empty' is False.")
                    % {"class_name": self.__class__.__name__}
                )

        return queryset

    def get_ordering(self):
        """Return the field or fields to use for ordering the queryset."""
        return self.ordering

    def get_paginate_by(self, queryset):
        """Get the number of items to paginate by, or ``None`` for no
        pagination."""
        return self.paginate_by

    def get_paginate_orphans(self):
        """Return the maximum number of orphans extend the last page by when
        paginating."""
        return self.paginate_orphans

    def get_paginator(self, queryset, per_page, orphans=0, allow_empty_first_page=True, **kwargs):
        """Return an instance of the paginator for this view."""
        return self.paginator_class(
            queryset, per_page, orphans=orphans, allow_empty_first_page=allow_empty_first_page, **kwargs
        )

    def paginate_queryset(self, queryset, page_size):
        """Paginate queryset."""
        paginator = self.get_paginator(
            queryset, page_size, orphans=self.get_paginate_orphans(), allow_empty_first_page=self.get_allow_empty()
        )
        page_kwarg = self.page_kwarg
        page = self.kwargs.get(page_kwarg) or self.request.GET.get(page_kwarg) or 1
        try:
            page_number = int(page)
        except ValueError:
            if page == "last":
                page_number = paginator.num_pages
            else:
                raise Http404(gettext("Page is not 'last', nor can it be converted to an int."))

        try:
            page = paginator.page(page_number)
            return (paginator, page, page.object_list, page.has_other_pages())

        except InvalidPage as e:
            raise Http404(
                gettext("Invalid page (%(page_number)s): %(message)s") % {"page_number": page_number, "message": str(e)}
            )

    def get_allow_empty(self):
        """Return ``True`` if the view should display empty lists and ``False``
        if a 404 should be raised instead."""
        return self.allow_empty


class BaseSingleObjectMixin(SQLAlchemyMixin):
    """Provides sqlalchemy support for detail views."""

    object = None
    slug_field = "slug"
    slug_url_kwarg = "slug"
    query_pkg_and_slug = False

    def get_object(self, queryset=None):
        """Return the object the view is displaying.

        Require `self.queryset` and the primary key attributes or the
        slug attributes in the URLconf. Subclasses can override this to
        return any object
        """
        if queryset is None:
            queryset = self.get_queryset()

        slug = self.kwargs.get(self.slug_url_kwarg)
        model = self.get_model()
        info = meta.model_info(model)
        pk = info.primary_keys_from_dict(self.kwargs)
        obj = None

        if pk is not None:
            obj = queryset.get(pk)

        if slug is not None and (pk is None and self.query_pkg_and_slug):
            slug_field = self.get_slug_field()
            obj = next(iter(queryset.filter_by(**{slug_field: slug})), None)

        if pk is None and slug is None:
            raise AttributeError(
                "Generic detail view %s must be called with either an object "
                "pk or a slug in the URLconf." % self.__class__.__name__
            )

        if obj is None:
            raise Http404(gettext("No {cls} instance found matching the query".format(cls=model.__name__)))

        return obj

    def get_slug_field(self):
        """Get the name of a slug field to be used to look up by slug."""
        return self.slug_field



import os

from django.test import TestCase, override_settings
from django_sorcery.db.url import (
    boolean,
    importable,
    importable_list,
    importable_list_tuples,
    integer,
    make_url,
    string,
    string_list,
)


@override_settings(
    SQLALCHEMY_CONNECTIONS={
        "bad": {},
        "minimal": {"DIALECT": "sqlite"},
        "from_env_preserve": {"DIALECT": "sqlite", "ALCHEMY_OPTIONS": {"foo": "bar"}},
        "default": {
            "DIALECT": "postgresql",
            "DRIVER": "psycopg2",
            "USER": "usr",
            "PASSWORD": "hunter2",
            "HOST": "pghost",
            "NAME": "pgdb",
            "PORT": "5432",
        },
        "ora": {"DIALECT": "oracle", "HOST": "orahost", "NAME": "oradb"},
    },
    DATABASES={
        "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"},
        "dummy": {"ENGINE": "django.db.backends.sqlite3", "DRIVER": "sqlite3", "NAME": "dummy.sqlite3"},
    },
)
class TestMakeUrl(TestCase):
    def setUp(self):
        super().setUp()
        os.environ["FROM_ENV_URL"] = "postgresql://usr:hunter2@awesomedomain/db?engine_echo=True"
        os.environ["FROM_ENV_PRESERVE_URL"] = "postgresql://usr:hunter2@awesomedomain/db?engine_echo=True"

    def tearDown(self):
        super().tearDown()
        os.environ.pop("FROM_ENV_URL", None)
        os.environ.pop("FROM_ENV_PRESERVE_URL", None)

    def test_boolean(self):
        self.assertTrue(boolean("True"))
        self.assertTrue(boolean("1"))
        self.assertFalse(boolean("False"))
        self.assertFalse(boolean("0"))

    def test_integer(self):
        self.assertEqual(integer("5"), 5)

    def test_string(self):
        self.assertEqual(string(b"hello"), "hello")

    def test_string_list(self):
        self.assertEqual(string_list(b"hello"), ["hello"])

    def test_importable(self):
        self.assertIs(importable("django.test.TestCase"), TestCase)
        self.assertIs(importable("os"), os)

    def test_importable_list(self):
        self.assertEqual(importable_list("django.test.TestCase"), [TestCase])

    def test_importable_list_tuples(self):
        self.assertEqual(importable_list_tuples("os:target"), [(os, "target")])

    def test_handles_url(self):
        url, options = make_url("postgresql://usr:hunter2@awesomedomain/db?engine_echo=True")
        self.assertEqual(url.database, "db")
        self.assertEqual(url.drivername, "postgresql")
        self.assertEqual(url.host, "awesomedomain")
        self.assertEqual(url.password, "hunter2")
        self.assertEqual(url.port, None)
        self.assertEqual(url.query, {})
        self.assertEqual(url.username, "usr")
        self.assertDictEqual(options, {"engine_options": {"echo": True}})

    def test_requires_dialect(self):
        with self.assertRaises(KeyError):
            make_url("bad")

    def test_can_override_from_env(self):
        url, options = make_url("from_env")
        self.assertEqual(url.database, "db")
        self.assertEqual(url.drivername, "postgresql")
        self.assertEqual(url.host, "awesomedomain")
        self.assertEqual(url.password, "hunter2")
        self.assertEqual(url.port, None)
        self.assertEqual(url.query, {})
        self.assertEqual(url.username, "usr")
        self.assertDictEqual(options, {"engine_options": {"echo": True}})

    def test_can_override_from_env_preserve(self):
        url, options = make_url("from_env_preserve")
        self.assertEqual(url.database, "db")
        self.assertEqual(url.drivername, "postgresql")
        self.assertEqual(url.host, "awesomedomain")
        self.assertEqual(url.password, "hunter2")
        self.assertEqual(url.port, None)
        self.assertEqual(url.query, {})
        self.assertEqual(url.username, "usr")
        self.assertDictEqual(options, {"engine_options": {"echo": True}, "foo": "bar"})

    def test_can_generate_minimal(self):
        url, _ = make_url("minimal")
        self.assertEqual(url.database, None)
        self.assertEqual(url.drivername, "sqlite")
        self.assertEqual(url.host, None)
        self.assertEqual(url.password, None)
        self.assertEqual(url.port, None)
        self.assertEqual(url.query, {})
        self.assertEqual(url.username, None)

    def test_can_generate_full(self):
        url, _ = make_url("default")
        self.assertEqual(url.database, "pgdb")
        self.assertEqual(url.drivername, "postgresql+psycopg2")
        self.assertEqual(url.host, "pghost")
        self.assertEqual(url.password, "hunter2")
        self.assertEqual(url.port, 5432)
        self.assertEqual(url.query, {})
        self.assertEqual(url.username, "usr")

    def test_can_handle_oracle_tns(self):
        url, _ = make_url("ora")
        self.assertEqual(url.database, "oradb")
        self.assertEqual(url.drivername, "oracle")
        self.assertEqual(url.host, "orahost")
        self.assertEqual(url.password, None)
        self.assertEqual(url.port, None)
        self.assertEqual(url.query, {})
        self.assertEqual(url.username, None)

    def test_can_generate_from_databases(self):
        url, _ = make_url("dummy")
        self.assertEqual(url.database, "dummy.sqlite3")
        self.assertEqual(url.drivername, "sqlite+sqlite3")
        self.assertEqual(url.host, None)
        self.assertEqual(url.password, None)
        self.assertEqual(url.port, None)
        self.assertEqual(url.query, {})
        self.assertEqual(url.username, None)
